"""A training pipeline which aims at being simple and modular. It is organized as follows:

1. Define three structures for training: Data, Model and Logger.
    - Data: a wrapper around train and val dataloaders.
    - Model: a wrapper containing:
        * a device (GPU or CPU basically)
        * a network
        * a decoder (optional)
        * an optimizer
        * a learning rate scheduler
        * a criterion (wrapper around an arbitrary number of criteria)
    - Logger: based on DataTracker, a convenient workaround to handle an arbitrary number of metrics
              and losses during training. It handles printing, tensorboard logging, and training
              state saving.

2. Define a Model factory function make_model handling:
    - training mode (PU or FS)
    - task (picking or tilt)
    - training config (num epochs, lr, backbone, swa, dim)
    - optionaly, reconstruction branch (autoencoder regularization)

3. Define functions to perform training and validation loop with logging.

In order to be used, one must simply instanciate a Data and Logger, call the make_model function,
and finally call the fit function accordingly.
That's what picking.py and tilt.py do.
"""

import functools
import os
import time
from collections import OrderedDict
from collections.abc import Iterable
from dataclasses import dataclass
from typing import Dict, List, Tuple

import numpy as np
import torch
from torch import Tensor
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter
from torch.utils.tensorboard.summary import hparams
from tqdm import tqdm

from .losses import BinomialGE, MultiTaskLoss
from .networks import Decoder, EfficientNet

# +------------------------------------------------------------------------------------------+ #
# |                                    TRAINING COMPONENTS                                   | #
# +------------------------------------------------------------------------------------------+ #


# ____________________________________________ DATA __________________________________________ #


@dataclass
class Data:
    """Nothing more than a wrapping structure handy for training functions."""

    train_loader: DataLoader
    val_loader: DataLoader


# ___________________________________________ MODEL __________________________________________ #


@dataclass
class Model:
    """Nothing more than a wrapping structure handy for training functions."""

    device: torch.device
    network: torch.nn.Module
    decoder: torch.nn.Module
    optimizer: torch.optim.Optimizer
    scheduler: torch.optim.lr_scheduler._LRScheduler
    train_criterion: MultiTaskLoss
    val_criterion: MultiTaskLoss


# ________________________________________ DATA TRACKER ______________________________________ #


class DataTracker:
    def __init__(self, names: List[str], init_best: float = 0.0) -> None:
        self.names = names
        self.total = 0
        self.current = {}
        self.best = self.init_best(init_best)
        self.reset()

    def init_best(self, value: float) -> None:
        best = {}
        for name in self.names:
            best[name] = value
        return best

    def step(self, *args, **kwargs) -> None:
        """Should update current and total."""
        pass

    def reduce(self) -> None:
        for name in self.names:
            self.current[name] /= self.total

    def reset(self) -> None:
        for name in self.names:
            self.current[name] = 0
        self.total = 0

    def update_best(self) -> None:
        for name in self.names:
            self.best[name] = self.current[name]


# _________________________________________ ACCURACY _________________________________________ #


class Accuracy(DataTracker):
    def __init__(self) -> None:
        super().__init__(["accuracy"], init_best=0)

    def step(self, outputs: torch.Tensor, targets: torch.Tensor) -> None:
        self.current["accuracy"] += outputs.argmax(dim=-1).eq(targets).sum().item()
        self.total += len(targets)


# __________________________________________ F1 SCORE ________________________________________ #


class F1Score(DataTracker):
    def __init__(self) -> None:
        super().__init__(["precision", "recall", "f1-score"], init_best=0)
        self.reset()

    def step(self, outputs: torch.Tensor, targets: torch.Tensor) -> None:
        predictions = outputs.detach().sigmoid().round()
        target_true = targets == 1
        predicted_true = predictions == 1
        self.correct_true += (
            (targets[target_true] == predictions[target_true]).sum().item()
        )
        self.target_true += target_true.sum().item()
        self.predicted_true += predicted_true.sum().item()

    def reduce(self) -> None:
        target_true_str = f"(target true: {self.target_true}, "
        predicted_true_str = f"predicted true: {self.predicted_true}, "
        correct_true_str = f"correct true: {self.correct_true})"
        print(target_true_str + predicted_true_str + correct_true_str)
        precision = (
            self.correct_true / self.predicted_true if self.predicted_true > 0 else 0
        )
        recall = self.correct_true / self.target_true if self.target_true > 0 else 0
        if precision + recall == 0:
            f1_score = 0
        else:
            f1_score = 2 * precision * recall / (precision + recall)
        self.current["precision"] = precision
        self.current["recall"] = recall
        self.current["f1-score"] = f1_score

    def reset(self) -> None:
        self.correct_true = 0
        self.target_true = 0
        self.predicted_true = 0


# ___________________________________________ LOSSES _________________________________________ #


class Losses(DataTracker):
    def __init__(self, names: List[str]) -> None:
        super().__init__(names, init_best=np.inf)

    def step(self, losses: List[Tensor], total_loss: Tensor, length: int) -> None:
        for i, name in enumerate(self.names[:-1]):  # exclude total
            self.current[name] += losses[i]
        self.current["total"] += total_loss
        self.total += length


# ___________________________________________ LOGGER _________________________________________ #


class SummaryWriter(SummaryWriter):
    """When logging hparams in addition to scalars, tensorboard will create a subfolder.
    This behavior allows to store multiple hparams config within the same run folder.
    However, it is not what we want here. This small workaround prevents that.
    From: https://github.com/pytorch/pytorch/issues/32651
    """

    def add_hparams(self, hparam_dict, metric_dict):
        torch._C._log_api_usage_once("tensorboard.logging.add_hparams")
        if not isinstance(hparam_dict, dict) or not isinstance(metric_dict, dict):
            raise TypeError("hparam_dict and metric_dict should be dictionary.")
        exp, ssi, sei = hparams(hparam_dict, metric_dict)
        logdir = self._get_file_writer().get_logdir()
        with SummaryWriter(log_dir=logdir) as w_hp:
            w_hp.file_writer.add_summary(exp)
            w_hp.file_writer.add_summary(ssi)
            w_hp.file_writer.add_summary(sei)
            for k, v in metric_dict.items():
                w_hp.add_scalar(k, v)


class Logger:
    def __init__(self, losses: Losses, metrics: DataTracker, config: Dict) -> None:
        self.losses = losses
        self.metrics = metrics
        self.config = config
        self.epoch = 0
        self.epochs = config["num_epochs"]
        self.log_dir = config["output_dir"]
        self.writer = SummaryWriter(log_dir=self.log_dir)
        self.prefix_length = self.set_prefix_length()

    def set_prefix_length(self) -> None:
        max_losses_length = max([len(name) for name in self.losses.names])
        max_metrics_length = max([len(name) for name in self.metrics.names])
        return max(max_losses_length, max_metrics_length)

    def step(
        self, losses: List[Tensor], total_loss: Tensor, outputs: Tensor, targets: Tensor
    ):
        self.losses.step(losses, total_loss, len(targets))
        self.metrics.step(outputs, targets)

    def reduce(self) -> None:
        self.losses.reduce()
        self.metrics.reduce()

    def reset(self) -> None:
        self.losses.reset()
        self.metrics.reset()

    def to_tensorboard(self, training: bool) -> None:
        stage = "Train" if training else "Validation"
        for name, loss in self.losses.current.items():
            self.writer.add_scalar(
                f"{name.capitalize()} Loss/{stage}", loss, self.epoch
            )
        for name, metric in self.metrics.current.items():
            self.writer.add_scalar(f"{name.capitalize()}/{stage}", metric, self.epoch)

    def hparams_and_metrics_to_tensorboard(self) -> None:
        metrics_dict = {f"hparams/{k}": v for k, v in self.metrics.best.items()}
        # tensorboard accepts tensor but not tuple or list ...
        for k, v in self.config.items():
            if isinstance(v, Iterable) and not isinstance(v, str):
                self.config[k] = torch.as_tensor(v)
        for k, v in metrics_dict.items():
            if isinstance(v, Iterable):
                metrics_dict[k] = torch.as_tensor(v)
        self.writer.add_hparams(self.config, metrics_dict)

    def save_training_state(self, model: Model) -> None:
        print("Saving new best model.")
        state = {
            "losses": self.losses.best,
            "metrics": self.metrics.best,
            "network_state_dict": model.network.state_dict(),
            "optimizer_state_dict": model.optimizer.state_dict(),
        }
        path = os.path.join(self.log_dir, "checkpoint.pt")
        torch.save(state, path)

    def print_dict(self, dict: Dict, prefix: str = None) -> None:
        for k, v in dict.items():
            print(f"{prefix} {k}{(self.prefix_length - len(k)) * '.'}: {v:.4f}")

    def on_full_epoch_start(self) -> None:
        print(f"\nEPOCH [{self.epoch + 1}/{self.epochs}]")

    def on_full_epoch_end(self, model: Model) -> None:
        self.epoch += 1
        lr = model.optimizer.param_groups[0]["lr"]
        self.writer.add_scalar("Learning Rate", lr, self.epoch)
        eta = model.train_criterion.eta
        for i, e in enumerate(eta):
            self.writer.add_scalar(f"Loss Weights/eta_{i}", e, self.epoch)
        self.print_dict(self.losses.current, prefix="Validation")
        self.print_dict(self.metrics.current, prefix="Validation")
        if self.losses.current["total"] <= self.losses.best["total"]:
            self.losses.update_best()
            self.metrics.update_best()
            self.save_training_state(model)

    def on_fit_end(self, model: Model) -> None:
        print(40 * "-")
        self.print_dict(self.losses.best, prefix="Best validation")
        self.print_dict(self.metrics.best, prefix="Best validation")
        print(f"\nNetwork num params: {model.network.num_params}")
        if model.decoder is not None:
            print(f"Decoder num params: {model.decoder.num_params}")
        print(40 * "-")
        self.hparams_and_metrics_to_tensorboard()


# +------------------------------------------------------------------------------------------+ #
# |                                       MODEL FACTORY                                      | #
# +------------------------------------------------------------------------------------------+ #


def make_model(
    # train params
    num_classes: str,
    dim: int,
    lr: float,
    patch_size: Tuple[int],
    num_epochs: int,
    pi: float,
    # model params
    network: str,
    width_factor: float,
    depth_factor: float,
    swa: bool,
    swa_lr: float,
    # autoencoder params
    ae: bool = False,
    eta: float = 0.0,
    learn_eta: bool = False,
    num_features: Tuple[int] = None,
    hidden_dim: int = None,
) -> Model:
    # 1. Device
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    # 2. Network
    kwargs = dict(
        dim=dim,
        num_classes=num_classes,
        width_factor=width_factor,
        depth_factor=depth_factor,
    )
    network = EfficientNet.from_name(network, **kwargs).to(device)
    params = [{"params": network.parameters()}]
    # 3. Classification loss
    # train and val losses are separated because in PU we don't use the same one
    # in train (GE Binomial) and in val (Binary Cross Entropy).
    if num_classes == 3:  # tilt
        train_criteria = [torch.nn.CrossEntropyLoss()]
        val_criteria = [torch.nn.CrossEntropyLoss()]
    else:  # picking
        # pi is None <=> FS or PU
        train_criteria = [
            BinomialGE(pi=pi) if pi is not None else torch.nn.BCEWithLogitsLoss()
        ]
        val_criteria = [torch.nn.BCEWithLogitsLoss()]
    # 4. Decoder and Reconstruction loss (optional)
    decoder = None
    if ae:
        decoder_kwargs = {
            "latent_dim": network.latent_dim,
            "hidden_dim": hidden_dim,
            "num_features": num_features,
            "dim": dim,
        }
        decoder = Decoder(out_size=patch_size, **decoder_kwargs).to(device)
        params.append({"params": decoder.parameters()})
        train_criteria.append(torch.nn.MSELoss())
        val_criteria.append(torch.nn.MSELoss())
    # 5. Merged loss
    multi_task_loss_kwargs = {
        "eta": len(train_criteria) * [eta],
        "learn_eta": learn_eta,
        "device": device,
    }
    train_criterion = MultiTaskLoss(train_criteria, **multi_task_loss_kwargs)
    if learn_eta:
        params.append({"params": train_criterion.parameters()})
    val_criterion = MultiTaskLoss(val_criteria, device=device)
    # 6. Optimizer
    adam_params = {
        "lr": lr,
        "betas": (0.9, 0.999),
        "weight_decay": 0.01,
        "amsgrad": False,
    }
    optimizer = torch.optim.AdamW(params, **adam_params)
    # sgd_params = {'lr': lr, 'momentum': 0.9, 'weight_decay': 5e-4}
    # optimizer = torch.optim.SGD(params, **sgd_params)
    # 7. Scheduler
    scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, T_max=num_epochs)
    # scheduler = None
    model = Model(
        device, network, decoder, optimizer, scheduler, train_criterion, val_criterion
    )
    # 8. Stochastic Weights Averaging (optional)
    model.swa = swa
    if model.swa:

        def ema_avg(averaged_model_parameter, model_parameter, num_averaged):
            return 0.1 * averaged_model_parameter + 0.9 * model_parameter

        ema_model = torch.optim.swa_utils.AveragedModel(model.network, avg_fn=ema_avg)
        model.swa_network = ema_model
        model.swa_scheduler = torch.optim.swa_utils.SWALR(
            model.optimizer, swa_lr=swa_lr
        )
        model.swa_start = num_epochs // 2
    return model


# +------------------------------------------------------------------------------------------+ #
# |                               TRAIN/VAL AUXILIARY FUNCTIONS                              | #
# +------------------------------------------------------------------------------------------+ #


def early_stopping(patience: int = 10):
    """Outer wrapper because the decorator has argument(s)."""

    def wrapper(epoch_function):
        epoch_function.current_run = 0
        """ The actual decorator. """

        @functools.wraps(epoch_function)
        def inner(*args, **kwargs):
            model, logger = epoch_function(*args, **kwargs)
            if logger.losses.current["total"] < logger.losses.best["total"]:
                epoch_function.current_run = 0  # reset patience
            else:
                epoch_function.current_run += 1
            if epoch_function.current_run == patience:
                print(f"Early stopping triggered (patience {patience}).")
                # will be used to detect early stopping in fit function
                logger.losses.current["total"] = None
            return model, logger

        return inner

    return wrapper


def timer(func):
    """A simple decorator to measure the execution time of a given function."""

    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        tic = time.perf_counter()
        value = func(*args, **kwargs)
        toc = time.perf_counter()
        elapsed_time = int(toc - tic)
        s = elapsed_time % 60
        m = (elapsed_time // 60) % 60
        h = elapsed_time // 3600
        formatted_elapsed_time = "{:d}h {:02d}m {:02d}s".format(h, m, s)
        print(f"Elapsed time: {formatted_elapsed_time}.")
        return value

    return wrapper_timer


def complete_swa(loader: DataLoader, model: Model, log_dir: str) -> None:
    print("Taking care of SWA batch norm statistics.")
    torch.optim.swa_utils.update_bn(loader, model.swa_network, model.device)
    print("Saving SWA weights.")
    # Make SWA modules names equal to original network ones.
    swa_state = model.swa_network.state_dict()
    #   - delete SWA specific key
    del swa_state["n_averaged"]
    #   - remove 'module.' prefix
    modules = [(".".join(k.split(".")[1:]), v) for k, v in swa_state.items()]
    #   - wrap results
    new_swa_state = OrderedDict(modules)
    # Save
    state = dict(network_state_dict=new_swa_state)
    torch.save(state, os.path.join(log_dir, "checkpoint_swa.pt"))


# +------------------------------------------------------------------------------------------+ #
# |                                  TRAIN/VAL MAIN FUNCTIONS                                | #
# +------------------------------------------------------------------------------------------+ #


def step(x: Tensor, y: Tensor, model: Model, logger: Logger, training: bool) -> None:
    x, y = x.to(model.device), y.to(model.device)
    z = model.network.extract_features(x)
    y_hat = model.network.head(z)
    outputs, targets = [y_hat], [y]
    if model.decoder is not None:
        model.val_criterion.eta = (
            model.train_criterion.eta
        )  # in case eta is being learnt
        x_hat = model.decoder(z)
        outputs.append(x_hat)
        targets.append(x)
    criterion = model.train_criterion if training else model.val_criterion
    losses, total_loss = criterion(outputs, targets)
    if training:
        model.optimizer.zero_grad()
        total_loss.backward()
        model.optimizer.step()
    losses = [loss.detach().item() for loss in losses]
    logger.step(losses, total_loss.detach().item(), y_hat.detach(), y)


def half_epoch(
    dataloader: DataLoader, model: Model, logger: Logger, training=True
) -> Model:
    model.network.train() if training else model.network.eval()
    torch.set_grad_enabled(training)
    for inputs, targets in tqdm(dataloader):
        step(inputs, targets, model, logger, training)
    logger.reduce()
    logger.to_tensorboard(training)
    return model


# @early_stopping(patience=10)
def full_epoch(data: Data, model: Model, logger: Logger) -> Tuple[Model, Logger]:
    model = half_epoch(data.train_loader, model, logger, training=True)
    logger.reset()
    if len(data.val_loader.dataset) > 0:
        model = half_epoch(data.val_loader, model, logger, training=False)
    else:
        print("Val data empty !")
    if model.swa and logger.epoch > model.swa_start:
        model.swa_network.update_parameters(model.network)
        model.swa_scheduler.step()
    elif model.scheduler is not None:
        model.scheduler.step()
    return model, logger


@timer
def fit(data: Data, model: Model, logger: Logger) -> None:
    for _ in range(logger.epochs):
        logger.on_full_epoch_start()
        model, logger = full_epoch(data, model, logger)
        if logger.losses.current["total"] is None:  # Early stopping
            break
        logger.on_full_epoch_end(model)
        logger.reset()
    logger.on_fit_end(model)
    if model.swa:
        complete_swa(data.train_loader, model, logger.log_dir)



import importlib.util


def has_cupy():
    return importlib.util.find_spec("cupy") is not None


def has_torch():
    return importlib.util.find_spec("torch") is not None


def has_torch_cuda():
    if has_torch():
        import torch

        if torch.cuda.is_available():
            return True
    return False



from typing import TYPE_CHECKING, TypeAlias

if TYPE_CHECKING:
    import array_api_strict as xp
    from array_api_strict._array_object import Array as ArrayAPIArray
    from array_api_strict._array_object import Device as ArrayAPIDevice
    from array_api_strict._array_object import Dtype as ArrayAPIDtype

    Array: TypeAlias = ArrayAPIArray
    Device: TypeAlias = ArrayAPIDevice
    Dtype: TypeAlias = ArrayAPIDtype

    array_api_module: TypeAlias = xp

def array_namespace(*xs) -> "array_api_module": ...
def get_prefered_namespace_device(
    xp: "array_api_module | None" = None,  # type: ignore
    device: "Device | None" = None,
    gpu: bool | None = None,
) -> "tuple[array_api_module, Device | None]": ...  # type: ignore
def median(a: "Array", axis: int | None, xp: "array_api_module"): ...



import numpy as np
from scipy.spatial.transform import Rotation as R

from spfluo import data
from spfluo.utils.transform import (
    distance_family_poses,
    distance_poses,
    get_transform_matrix,
    symmetrize_poses,
)
from spfluo.utils.volume import affine_transform, are_volumes_aligned


def test_get_transform_matrix_simple():
    H = get_transform_matrix(
        (30, 30, 30),
        np.array([1.0, 2.0, 3.0]),
        np.array([4.0, 5.0, 6.0]),
        convention="ZXZ",
        degrees=False,
    )
    expected_result = np.array(
        [
            [-0.48547846, -0.42291857, 0.7651474, 20.57711966],
            [-0.8647801, 0.10384657, -0.4912955, 37.65732099],
            [0.12832006, -0.90019763, -0.41614684, 37.72635389],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )
    assert np.isclose(H, expected_result).all()


def test_get_transform_matrix_batch():
    output_shape = (30, 30, 30)
    N = 10
    rot = np.random.randn(N, 3)
    trans = np.random.randn(N, 3)

    matrices = get_transform_matrix(output_shape, rot, trans)
    for i in range(N):
        assert np.isclose(
            matrices[i], get_transform_matrix(output_shape, rot[i], trans[i])
        ).all()


def test_distance_poses():
    p1 = np.asarray([90, 90, 0, 1, 0, 0], dtype=float)
    p2 = np.asarray([-90, 90, 0, 0, 1, 0], dtype=float)
    angle, t = distance_poses(p1, p2)

    assert np.isclose(angle, 180.0)
    assert np.isclose(t, 2.0**0.5)


def test_distance_family_poses():
    p1 = np.asarray(
        [[70, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0], [90, 90, 0, 1, 0, 0]], dtype=float
    )
    p2 = np.asarray(
        [[70, 0, 0, 0, 0, 0], [90, 0, 0, 0, 0, 0], [90, 90, 0, 0, 1, 0]], dtype=float
    )
    angle, t = distance_family_poses(p1, p2)
    assert np.isclose(angle, [0, 90, 0], atol=1e-5).all()
    assert np.isclose(t, [0, 0, 2.0**0.5]).all()


def test_distance_family_poses_sym():
    symmetry = 9
    error = 0.2 * (360 / symmetry)
    p1 = np.asarray(
        [
            [70, 0, 0, 0, 0, 0],
            [5 * 360 / symmetry + error, 0, 0, 0, 0, 0],
            [90, 90, 0, 1, 0, 0],
        ],
        dtype=float,
    )
    p2 = np.asarray(
        [[70, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0], [90, 90, 0, 0, 1, 0]], dtype=float
    )
    angle, t = distance_family_poses(p1, p2, symmetry=9)
    assert np.isclose(angle, [0, error, 0], atol=1e-5).all()
    assert np.isclose(t, [0, 0, 2.0**0.5]).all()


def test_distance_family_poses_rotoffset():
    p2 = np.asarray(
        [[70, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0], [90, 90, 0, 0, 0, 0]], dtype=float
    )
    R_p2 = R.from_euler("XZX", p2[:, :3], degrees=True)
    R0 = R.random(random_state=0)
    R_p1 = R_p2 * R0
    p1 = np.zeros_like(p2)
    p1[:, :3] = R_p1.as_euler("XZX", degrees=True)
    angle, t = distance_family_poses(p1, p2)
    assert np.isclose(angle, [0, 0, 0], atol=1e-5).all()
    assert np.isclose(t, [0, 0, 0]).all()


def test_distance_family_poses_rotoffset_sym():
    p2 = np.asarray(
        [[70, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0], [90, 90, 0, 0, 0, 0]], dtype=float
    )
    R_p2 = R.from_euler("XZX", p2[:, :3], degrees=True)
    R0 = R.random(random_state=0)
    R_p1 = R_p2 * R0
    p1 = np.zeros_like(p2)
    p1[:, :3] = R_p1.as_euler("XZX", degrees=True)
    angle, t = distance_family_poses(p1, p2)
    assert np.isclose(angle, [0, 0, 0], atol=1e-5).all()
    assert np.isclose(t, [0, 0, 0]).all()


def test_symmetrize_poses():
    poses, volume = (
        data.generated_isotropic()["poses"],
        data.generated_isotropic()["volumes"][0],
    )
    gt = data.generated_isotropic()["gt"]
    poses_sym = symmetrize_poses(poses, 9)
    mats = get_transform_matrix(
        volume.shape, poses_sym[0, :, :3], poses_sym[0, :, 3:], degrees=True
    )
    volume_rotated = affine_transform(np.stack((volume,) * 9), mats, batch=True)
    assert are_volumes_aligned(volume_rotated, gt).all()



# Make the generated data available to all spfluo subpackages
import functools
import os
from pathlib import Path
from typing import TYPE_CHECKING, Tuple, Union

import numpy as np
import pytest
import tifffile
from hypothesis import settings

from spfluo import data
from spfluo.utils.array import to_numpy

settings.register_profile("dev", max_examples=10)
settings.load_profile(os.getenv("HYPOTHESIS_PROFILE", "dev"))

if TYPE_CHECKING:
    from spfluo.utils.array import Array


def pytest_addoption(parser: pytest.Parser):
    parser.addoption("--image-directory", action="store", default=None)


@pytest.fixture(scope="session")
def image_directory(pytestconfig: pytest.Config):
    dir = pytestconfig.getoption("--image-directory")
    if dir:
        dir = Path(dir)
        if not dir.exists():
            dir.mkdir()
    else:
        dir = None
    return dir


@pytest.fixture(scope="function")
def save_result(image_directory: Union[Path, None], request: pytest.FixtureRequest):
    base_name = os.path.split(request.node.nodeid)[1]

    def inner(name: str, arr: "Array", *args, **kwargs):
        saved = False
        if image_directory:
            tifffile.imwrite(
                str(image_directory / (base_name + "-" + name + ".ome.tiff")),
                to_numpy(arr),
                *args,
                **kwargs,
            )
            saved = True
        return saved

    return functools.wraps(tifffile.imwrite)(inner)


@pytest.fixture(
    scope="session",
    params=[data.generated_isotropic(), data.generated_anisotropic()],
    ids=["isotropic", "anisotropic"],
)
def generated_data_all(
    request,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    return tuple(request.param[k] for k in ["volumes", "poses", "psf", "gt"])


@pytest.fixture(scope="session")
def generated_data_anisotropic() -> (
    Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
):
    d = data.generated_anisotropic()
    return tuple(d[k] for k in ["volumes", "poses", "psf", "gt"])



from __future__ import annotations

import logging
import math
from functools import partial
from typing import TYPE_CHECKING, Optional

import numpy as np
import tifffile
from ome_types import from_xml
from ome_types.model.simple_types import UnitsLength
from sklearn.decomposition import PCA

from spfluo.refinement.refinement import create_poses_grid
from spfluo.utils.array import (
    array_namespace,
    get_device,
    is_cupy_array,
    is_numpy_array,
    to_numpy,
)
from spfluo.utils.loading import get_data_from_ome_tiff, read_poses, save_poses
from spfluo.utils.transform import (
    compose_poses,
    get_transform_matrix_around_center,
    get_transform_matrix_from_pose,
    invert_pose,
    matrix_to_euler,
)
from spfluo.utils.volume import affine_transform, interpolate_to_size, resample

logger = logging.getLogger("spfluo.utils.rotate_symmetry_axis")

if TYPE_CHECKING:
    from spfluo.utils.array import Array
DEFAULT_THRESHOLD = 0.3


def convert_im_to_point_cloud(im: "Array", thesh: float):
    xp = array_namespace(im)
    coordinates = xp.nonzero(im >= thesh)
    coordinates = xp.stack(coordinates, axis=-1)
    return coordinates


def skew_symmetric_cross_product(v: "Array"):
    v1, v2, v3 = v[0], v[1], v[2]
    xp = array_namespace(v)
    return xp.asarray(
        [[0, -v3, v2], [v3, 0, -v1], [-v2, v1, 0]], dtype=v.dtype, device=get_device(v)
    )


def find_rotation_between_two_vectors(a: "Array", b: "Array"):
    """returns the rotation matrix that rotates vector a onto vector b
    (the rotation matrix s.t. Ra = b)"""
    xp = array_namespace(a, b)
    device = get_device(a)
    dtype = a.dtype
    v = xp.linalg.cross(a, b)
    s = xp.linalg.vector_norm(v)
    if not (s > 0):
        # a and b are colinear
        x, y, z = a
        if x > 0:
            v = xp.asarray([y, -x, 0])
        elif y > 0:
            v = xp.asarray([y, -x, 0])
        else:  # x == y == 0, a is colinear to z-axis
            v = xp.asarray([0, 0, 1])
        s = xp.linalg.vector_norm(v)
    c = xp.linalg.vecdot(a, b)
    ssc = skew_symmetric_cross_product(v)
    rot = xp.eye(3, 3, dtype=dtype, device=device) + ssc + (ssc @ ssc) * (1 - c) / s**2
    return rot


def threshold_otsu(
    image: "Array | None " = None,
    nbins: int | None = 256,
    *,
    hist: "Array | None" = None,
) -> "Array":
    xp = array_namespace(image)
    if is_cupy_array(image):
        from cucim.skimage.filters import threshold_otsu as threshold_otsu_cucim

        func = threshold_otsu_cucim
    elif is_numpy_array(image):
        from skimage.filters import threshold_otsu as threshold_otsu_skimage

        func = threshold_otsu_skimage
    else:
        raise NotImplementedError(f"threshold_otsu is not implemented for {xp} backend")

    return func(image, nbins, hist=hist)


def find_pose_from_z_axis_to_centriole_axis2(
    centriole_im: "Array", axis_indice: int = 0, convention="XZX"
):
    """New method to get the centriole axis
    We test all possible poses in SO3. We evaluate if the pose
    is good by projecting the volume rotated on the Z axis, thresholding it
    and looking at the total surface.
    If the surface is small we probably have the centriole seen from the top,
    if not the pose is probably wrong.
    """
    xp = array_namespace(centriole_im)
    device = get_device(centriole_im)
    M = 1000
    potential_poses, precision = create_poses_grid(
        xp, M, 1, device=device, dtype=xp.float64
    )

    H = get_transform_matrix_from_pose(
        centriole_im.shape,
        potential_poses,
        convention=convention,
    )

    volume_rotated = affine_transform(
        xp.stack((centriole_im,) * M),
        H,
        order=1,
        prefilter=False,
        batch=True,
        multichannel=False,
    )
    volume_rotated_proj = xp.sum(volume_rotated, axis=1 + axis_indice)
    thresh = xp.stack(
        [
            threshold_otsu(volume_rotated_proj[i])
            for i in range(volume_rotated_proj.shape[0])
        ]
    )[:, None, None]

    best = xp.argmin(
        xp.sum((volume_rotated_proj > thresh), axis=(1, 2)),
    )
    return potential_poses[best]


def find_centriole_symmetry_axis(
    centriole_im: "Array", threshold: float = DEFAULT_THRESHOLD
):
    xp = array_namespace(centriole_im)
    device = get_device(centriole_im)
    ma = xp.max(centriole_im)
    centriol_pc = convert_im_to_point_cloud(centriole_im, ma * threshold)

    # PCA in numpy
    pca = PCA(n_components=3)
    pca.fit(to_numpy(centriol_pc))
    sum_of_2_by_2_differences = np.zeros(3)
    for i in range(3):
        for j in range(3):
            if j != i:
                sum_of_2_by_2_differences[i] += np.abs(
                    pca.singular_values_[i] - pca.singular_values_[j]
                )
    idx_dim_pca = np.argmax(sum_of_2_by_2_differences)
    logger.debug(f"{pca.singular_values_=}")
    logger.debug(f"{pca.components_=}")
    logger.debug(f"{sum_of_2_by_2_differences=}")
    if logger.isEnabledFor(logging.DEBUG):
        import plotly.graph_objects as go

        from spfluo.utils.pointcloud_viz import interactive_plot

        center = np.mean(centriol_pc, axis=0)
        fig = interactive_plot([centriol_pc], return_fig=True)
        for i, comp in enumerate(pca.components_):
            fig.add_trace(
                go.Scatter3d(
                    x=[center[0], center[0] + comp[0]],
                    y=[center[1], center[1] + comp[1]],
                    z=[center[2], center[2] + comp[2]],
                    mode="lines",
                    line=dict(color="red", width=3),
                    name=f"Composante {i+1}",
                )
            )
        fig.show()

    # return a generic Array
    return xp.asarray(
        pca.components_[idx_dim_pca], device=device, dtype=xp.float64
    ), xp.asarray(pca.mean_, device=device, dtype=xp.float64)


def find_pose_from_z_axis_to_centriole_axis(
    centriole_im: "Array", axis_indice=0, threshold=DEFAULT_THRESHOLD, convention="XZX"
):
    """Find the pose of the transformation from the axis to the centriole"""
    xp = array_namespace(centriole_im)
    device = get_device(centriole_im)
    symmetry_axis, center = find_centriole_symmetry_axis(
        centriole_im, threshold=threshold
    )
    z_axis = xp.asarray([0, 0, 0], device=device, dtype=xp.float64)
    z_axis[axis_indice] = 1.0
    rot = find_rotation_between_two_vectors(symmetry_axis, z_axis)
    pose = xp.zeros((6,), dtype=xp.float64)
    pose[:3] = matrix_to_euler(convention, rot.T, degrees=True)
    return xp.asarray(pose, dtype=xp.float64, device=device)


def _gradient(im: "Array", nb_spatial_dims: int = 2):
    xp = array_namespace(im)
    grad = []
    for i in range(-nb_spatial_dims, 0):
        selection_a, selection_b = [], []
        for j in range(-im.ndim, 0):
            if j < -nb_spatial_dims:
                selection_a.append(slice(None))
                selection_b.append(slice(None))
            elif j == i:
                selection_a.append(slice(1, None))
                selection_b.append(slice(None, -1))
            else:
                selection_a.append(slice(None, -1))
                selection_b.append(slice(None, -1))
        grad_i = im[tuple(selection_a)] - im[tuple(selection_b)]
        grad.append(grad_i)
    return xp.stack(grad, axis=0)


def _corr(im1: "Array", im2: "Array", nb_spatial_dims: int = 2):
    xp = array_namespace(im1, im2)
    assert im1.shape[-nb_spatial_dims:] == im2.shape[-nb_spatial_dims:]
    mean = partial(xp.mean, axis=tuple(range(-nb_spatial_dims, 0)), keepdims=True)
    im1_stdv = mean((im1 - mean(im1)) ** 2) ** 0.5
    im2_stdv = mean((im2 - mean(im2)) ** 2) ** 0.5
    c = mean((im1 - mean(im1)) * (im2 - mean(im2))) / (im1_stdv * im2_stdv)
    c = c[
        (slice(None),) * (c.ndim - nb_spatial_dims) + (0,) * nb_spatial_dims
    ]  # squeeze dims
    return c


def find_pose_from_centriole_to_center(
    im: "Array", symmetry: int, precision: float = 1, axis_indice: int = 0
):
    xp = array_namespace(im)
    dtype = im.dtype
    device = get_device(im)
    # Z-trans
    axes = [0, 1, 2]
    axes.pop(axis_indice)
    axes = tuple(axes)
    center_mass = xp.sum(
        xp.arange(im.shape[axis_indice], device=get_device(im)) * xp.sum(im, axis=axes)
    ) / xp.sum(im)
    center = (float(im.shape[axis_indice]) - 1) / 2
    trans_z = center - center_mass

    # Y-X trans
    num = math.ceil(max(im.shape[1:]) / (4 * precision))
    N_trans = num * num
    im_proj = xp.sum(im, axis=axis_indice)
    yx_translations = xp.reshape(
        xp.stack(
            xp.meshgrid(xp.linspace(-1, 1, num), xp.linspace(-1, 1, num)), axis=-1
        ),
        (num * num, 2),
    )
    yx_translations = yx_translations * xp.asarray(im_proj.shape) / 4

    H_trans = xp.stack((xp.eye(3),) * N_trans)
    H_trans[:, :2, 2] = yx_translations

    angles = 2 * xp.arange(symmetry, dtype=dtype) * xp.pi / symmetry
    R = xp.permute_dims(
        xp.asarray(
            xp.stack(
                [
                    xp.stack([xp.cos(angles), -xp.sin(angles)]),
                    xp.stack([xp.sin(angles), xp.cos(angles)]),
                ]
            )
        ),
        axes=(2, 0, 1),
    )
    H_rot = get_transform_matrix_around_center(im_proj.shape, R)

    H = H_rot[None] @ H_trans[:, None]  # (N_trans, symmetry, 3, 3)
    H_inv = xp.linalg.inv(H)

    H_inv[:, :, 2, [0, 1, 2]] = xp.asarray(
        [0.0, 0.0, 1.0], dtype=H_inv.dtype, device=get_device(H_inv)
    )
    ims_translated_rotated = xp.reshape(
        xp.asarray(
            affine_transform(
                to_numpy(
                    xp.stack((im_proj,) * N_trans * symmetry)
                ),  # to numpy because affine transform only works in numpy for 2D
                to_numpy(xp.reshape(H_inv, (-1, 3, 3))),
                batch=True,
            ),
            device=device,
            dtype=dtype,
        ),
        (N_trans, symmetry, *im_proj.shape),
    )

    corr_rotations = xp.mean(
        _corr(
            xp.mean(ims_translated_rotated, axis=1, keepdims=True),
            ims_translated_rotated,
        ),
        axis=1,
    )

    grad = _gradient(ims_translated_rotated)
    corr_rotations_grads = xp.mean(
        _corr(xp.mean(grad, axis=2, keepdims=True), grad), axis=(0, 2)
    )

    distances = (corr_rotations + corr_rotations_grads) / 2

    trans_y, trans_x = yx_translations[xp.argmax(distances)]

    pose = xp.asarray([0, 0, 0, 0, 0, 0], dtype=xp.float64, device=device)
    pose[3 + axis_indice] = trans_z
    pose[3 + axes[0]] = trans_y
    pose[3 + axes[1]] = trans_x

    return pose


def find_pose_from_z_axis_centered_to_centriole_axis(
    centriole_im: "Array",
    symmetry: int,
    axis_indice=0,
    threshold: float = DEFAULT_THRESHOLD,
    center_precision: float = 1,
    convention="XZX",
):
    logger.info("start find_pose_from_z_axis_centered_to_centriole_axis")
    pose_from_z_axis_to_centriole = find_pose_from_z_axis_to_centriole_axis(
        centriole_im,
        axis_indice=axis_indice,
        threshold=threshold,
        convention=convention,
    )
    logger.debug(f"{pose_from_z_axis_to_centriole=}")
    volume_z_axis = affine_transform(
        centriole_im,
        get_transform_matrix_from_pose(
            centriole_im.shape, pose_from_z_axis_to_centriole, convention=convention
        ),
        order=1,
    )
    pose_from_z_axis_to_z_axis_centered = find_pose_from_centriole_to_center(
        volume_z_axis, symmetry, axis_indice=axis_indice, precision=center_precision
    )
    logger.debug(f"{pose_from_z_axis_to_z_axis_centered=}")
    return compose_poses(
        invert_pose(pose_from_z_axis_to_z_axis_centered), pose_from_z_axis_to_centriole
    )


def main(
    volume_path: str,
    symmetry: int,
    convention: str = "XZX",
    threshold: float = DEFAULT_THRESHOLD,
    output_volume_path: Optional[str] = None,
    poses_path: Optional[str] = None,
    output_poses_path: Optional[str] = None,
):
    tif = tifffile.TiffFile(volume_path, is_ome=True)
    volume = get_data_from_ome_tiff(tif, 0, order="CZYX")

    # Volume is mono-channel
    assert volume.shape[0] == 1, f"volume of shape {volume.shape} is not monochannel"
    volume = volume[0]

    ome = from_xml(tif.ome_metadata)
    assert len(ome.images) == 1

    # if volume has pixel sizes, resample it to be isotropic
    if (
        ome.images[0].pixels.physical_size_x
        and ome.images[0].pixels.physical_size_y
        and ome.images[0].pixels.physical_size_z
    ):
        # assert all physical sizes have the same unit
        assert (
            len(
                set(
                    [
                        ome.images[0].pixels.physical_size_x_unit,
                        ome.images[0].pixels.physical_size_y_unit,
                        ome.images[0].pixels.physical_size_z_unit,
                    ]
                )
            )
            == 1
        )
        target_pixel_physical_unit = ome.images[0].pixels.physical_size_x_unit
        target_pixel_physical_size = min(
            ome.images[0].pixels.physical_size_z,
            ome.images[0].pixels.physical_size_y,
            ome.images[0].pixels.physical_size_x,
        )

        volume = resample(
            volume,
            (
                ome.images[0].pixels.physical_size_z / target_pixel_physical_size,
                ome.images[0].pixels.physical_size_y / target_pixel_physical_size,
                ome.images[0].pixels.physical_size_x / target_pixel_physical_size,
            ),
        )
    else:
        target_pixel_physical_size = 1.0
        target_pixel_physical_unit = UnitsLength.MICROMETER

    # Volume must be a cube
    size = max(volume.shape)
    volume = interpolate_to_size(volume, (size, size, size))

    pose = find_pose_from_z_axis_centered_to_centriole_axis(
        volume,
        symmetry,
        threshold=threshold,
    )

    if output_volume_path:
        rotated_volume = affine_transform(
            volume,
            get_transform_matrix_from_pose(volume.shape, pose, convention=convention),
        )
        assert output_volume_path.endswith(".ome.tiff")
        tifffile.imwrite(
            output_volume_path,
            rotated_volume,
            metadata={
                "axes": "ZYX",
                "PhysicalSizeX": target_pixel_physical_size,
                "PhysicalSizeXUnit": target_pixel_physical_unit.value,
                "PhysicalSizeY": target_pixel_physical_size,
                "PhysicalSizeYUnit": target_pixel_physical_unit.value,
                "PhysicalSizeZ": target_pixel_physical_size,
                "PhysicalSizeZUnit": target_pixel_physical_unit.value,
            },
        )
    if poses_path:
        poses, names = read_poses(poses_path)
        new_poses = compose_poses(pose, poses, convention=convention)
        save_poses(output_poses_path, new_poses, names)



from pathlib import Path

import numpy as np

from spfluo.picking.modules.pretraining.generate.config import (
    IO,
    Augmentation,
    DataGenerationConfig,
    Sensor,
    Voxelisation,
)
from spfluo.picking.modules.pretraining.generate.data_generator import DataGenerator

DATA_DIR = Path(__file__).parent

if __name__ == "__main__":
    np.random.seed(123)

    # Create Data Generator configs
    configs: list[DataGenerationConfig] = []
    ids: list[str] = []

    def default_config():
        D = 50
        N = 10
        pointcloud_path = DATA_DIR / "sample_centriole_point_cloud.csv"
        return DataGenerationConfig(
            dtype=np.float32,
            io=IO(
                point_cloud_path=pointcloud_path,
                extension="tiff",
            ),
            voxelisation=Voxelisation(
                num_particles=N,
                max_particle_dim=int(0.6 * D),
                image_shape=D,
                bandwidth=17,
            ),
            sensor=Sensor(gaussian_noise_target_snr_db=25),
            augmentation=Augmentation(rotation_proba=1, shrink_range=(1.0, 1.0)),
        )

    # isotropic / anisotropic datasets
    for anisotropic_param, id in [
        ((1.0, 1.0, 1.0), "isotropic-1.0"),
        ((5.0, 1.0, 1.0), "anisotropic-5.0-1.0-1.0"),
        ((10.0, 1.0, 1.0), "anisotropic-10.0-1.0-1.0"),
    ]:
        config = default_config()
        config.augmentation.max_translation = 0
        config.sensor.anisotropic_blur_sigma = anisotropic_param
        if anisotropic_param == (10.0, 1.0, 1.0):
            config.voxelisation.image_shape = 70

        configs.append(config)
        ids.append(id)

    # dataset with translations
    config = default_config()
    config.augmentation.max_translation = 100
    configs.append(config)
    ids.append("with-translations")

    # Generate datasets
    for config, id in zip(configs, ids):
        output_dir = DATA_DIR / "generated" / id
        output_dir.mkdir(exist_ok=True, parents=True)
        gen = DataGenerator(config)
        gt_path = output_dir / "gt.tiff"
        gen.save_psf(output_dir / "psf.tiff")
        gen.save_groundtruth(gt_path)
        gen.create_particles(output_dir, output_extension="tiff")



from typing import TYPE_CHECKING

import pytest
from hypothesis import given, settings
from hypothesis import strategies as st
from hypothesis.extra.numpy import arrays
from scipy.ndimage import affine_transform as affine_transform_scipy
from scipy.ndimage import fourier_shift as fourier_shift_scipy
from scipy.ndimage import shift as shift_scipy
from skimage import data, util
from skimage.registration import (
    phase_cross_correlation as phase_cross_correlation_skimage,
)

from spfluo.tests.helpers import assert_allclose, random_pose, testing_libs
from spfluo.utils.array import numpy as np
from spfluo.utils.array import to_numpy
from spfluo.utils.transform import get_transform_matrix_from_pose
from spfluo.utils.volume import affine_transform, fourier_shift, phase_cross_correlation

if TYPE_CHECKING:
    from spfluo.utils.array import array_api_module


##################################################################
# Test phase_cross_correlation against the scikit-image function #
##################################################################


@settings(deadline=None)
@given(
    translation=arrays(
        float,
        (3,),
        elements=st.floats(
            min_value=-10, max_value=10, allow_nan=False, allow_infinity=False
        ),
    ),
    upsample_factor=st.integers(min_value=1, max_value=100),
    space=st.sampled_from(["fourier", "real"]),
    normalization=st.sampled_from(["phase", None]),
    half_precision=st.sampled_from([True, False]),
)
@pytest.mark.parametrize(
    "xp, device",
    testing_libs,
)
@pytest.mark.parametrize("image", [data.camera(), data.cells3d()[:, 0, :60, :60]])
def test_correctness_phase_cross_correlation(
    xp: "array_api_module",
    device,
    image,
    translation,
    upsample_factor,
    space,
    normalization,
    half_precision,
):
    translation = translation[: image.ndim]
    if space == "fourier":
        reference_image = np.fft.fftn(util.img_as_float(image))
        moving_image = fourier_shift(reference_image, translation)
        if half_precision:
            reference_image = reference_image.astype(np.complex64)
            moving_image = moving_image.astype(np.complex64)
    else:
        reference_image = util.img_as_float(image)
        moving_image = shift_scipy(reference_image, translation)
        if half_precision:
            reference_image = reference_image.astype(np.float32)
            moving_image = moving_image.astype(np.float32)

    reference_image_xp = xp.asarray(reference_image, device=device)
    moving_image_xp = xp.asarray(moving_image, device=device)

    (shift, error, phasediff), (shift_skimage, error_skimage, phasediff_skimage) = [
        func(
            reference_image_,
            moving_image_,
            space=space,
            upsample_factor=upsample_factor,
            normalization=normalization,
        )
        for func, reference_image_, moving_image_ in [
            (phase_cross_correlation, reference_image_xp, moving_image_xp),
            (phase_cross_correlation_skimage, reference_image, moving_image),
        ]
    ]
    translation = to_numpy(translation)
    shift_skimage = -np.asarray(to_numpy(*shift_skimage))
    shift = -np.asarray(to_numpy(*shift))
    norm_error_vec_skimage = np.linalg.vector_norm(translation - shift_skimage)
    norm_error_vec = np.linalg.vector_norm(translation - shift)
    if norm_error_vec_skimage < norm_error_vec:
        assert_allclose(
            norm_error_vec, norm_error_vec_skimage, rtol=0.1, atol=1 / upsample_factor
        )  # atol=0.1
    assert_allclose(error, xp.asarray(error_skimage), rtol=1e-3, atol=1e-2)
    # Phasediff not tested because not implemented in pytorch


def test_broadcasting_phase_cross_correlation():  # TODO
    pass


@pytest.fixture
def cells3d_translated():
    image = np.transpose(
        data.cells3d()[:, :, :60, :60], (1, 0, 2, 3)
    )  # shape (C, D, H, W)
    translation = (5, -4, 10)

    reference_image = np.fft.fftn(util.img_as_float(image), axes=(1, 2, 3))
    moving_image = fourier_shift(reference_image, np.asarray(translation))
    return reference_image, moving_image, translation


@pytest.mark.parametrize("xp, device", testing_libs)
def test_multichannel_phase_cross_correlation_1(
    xp: "array_api_module", device, cells3d_translated
):
    reference_image, moving_image, translation = cells3d_translated
    reference_image_xp = xp.asarray(reference_image, device=device)
    moving_image_xp = xp.asarray(moving_image, device=device)

    shift, error, phasediff = phase_cross_correlation(
        reference_image_xp,
        moving_image_xp,
        space="fourier",
        nb_spatial_dims=3,
        multichannel=True,
    )
    assert len(shift) == 3
    for i in range(len(shift)):
        assert shift[i].ndim == 0
    assert error.ndim == 0
    assert_allclose(xp.stack(shift), -xp.asarray(translation))
    assert_allclose(error, xp.asarray(1.0), rtol=1e-5)


# With zeros on first channel
@pytest.mark.parametrize("xp, device", testing_libs)
def test_multichannel_phase_cross_correlation_2(
    xp: "array_api_module", device, cells3d_translated
):
    reference_image, moving_image, translation = cells3d_translated
    reference_image_xp = xp.asarray(reference_image, device=device)
    moving_image_xp = xp.asarray(moving_image, device=device)
    moving_image_xp_ = xp.asarray(moving_image_xp, copy=True)
    moving_image_xp_[0, ...] = 0.0
    shift, error, phasediff = phase_cross_correlation(
        reference_image_xp,
        moving_image_xp_,
        space="fourier",
        nb_spatial_dims=3,
        multichannel=True,
    )
    assert len(shift) == 3
    for i in range(len(shift)):
        assert shift[i].ndim == 0
    assert error.ndim == 0
    assert_allclose(xp.stack(shift), -xp.asarray(translation))
    assert_allclose(error, xp.asarray(1.0), rtol=1e-5)


# With upsampling
@pytest.mark.parametrize("xp, device", testing_libs)
def test_multichannel_phase_cross_correlation_upsampling(
    xp: "array_api_module", device
):
    image = np.transpose(
        data.cells3d()[:, :, :60, :60], (1, 0, 2, 3)
    )  # shape (C, D, H, W)
    translation = (5.3, -4.1, 10.4)

    reference_image = np.fft.fftn(util.img_as_float(image), axes=(1, 2, 3))
    moving_image = fourier_shift(reference_image, np.asarray(translation))

    reference_image_xp = xp.asarray(reference_image, device=device)
    moving_image_xp = xp.asarray(moving_image, device=device)
    shift, error, phasediff = phase_cross_correlation(
        reference_image_xp,
        moving_image_xp,
        space="fourier",
        nb_spatial_dims=3,
        multichannel=True,
        upsample_factor=10,
    )
    assert len(shift) == 3
    for i in range(len(shift)):
        assert shift[i].ndim == 0
    assert error.ndim == 0
    assert_allclose(xp.stack(shift), -xp.asarray(translation))
    assert_allclose(error, xp.asarray(1.0), rtol=1e-5)


####################################################
# Test affine_transform against the scipy function #
####################################################


def is_affine_close(im1, im2):
    """Scipy's and pytorch interpolations at borders don't behave equivalently
    So we add a margin"""
    D, H, W = im1.shape
    return np.isclose(im1, im2).sum() > (D * W * H - 2 * (H * D + D * W + W * H))


@settings(deadline=None)  # necessary for GPU testing
@given(
    pose=random_pose(),
    order=st.just(1),
    mode=st.sampled_from(["constant", "nearest", "reflect"]),
    cval=st.just(0.0),
    prefilter=st.just(True),
    d=st.data(),
)
@pytest.mark.parametrize(
    "xp, device",
    testing_libs,
)
@pytest.mark.parametrize("input", [util.img_as_float(data.cells3d()[:, 0, :60, :60])])
def test_correctness_affine_transform(
    xp: "array_api_module", input, device, pose, order, mode, cval, prefilter, d
):
    output_shape = d.draw(
        st.tuples(
            *[
                st.integers(min_value=input.shape[i] // 2, max_value=input.shape[i] * 2)
                for i in range(input.ndim)
            ],
        )
    )
    matrix = get_transform_matrix_from_pose(output_shape, pose)
    input_xp = xp.asarray(input, device=device)
    matrix_xp = xp.asarray(matrix, device=device)

    output_xp, output = [
        func(
            input_,
            matrix_,
            output_shape=output_shape,
            order=order,
            mode=mode,
            cval=cval,
            prefilter=prefilter,
        )
        for func, input_, matrix_ in [
            (affine_transform, input_xp, matrix_xp),
            (affine_transform_scipy, input, matrix),
        ]
    ]

    assert is_affine_close(to_numpy(output_xp), output)


#################################################
# Test fourier_shift against the scipy function #
#################################################


@settings(deadline=None)
@given(
    shift=arrays(
        float,
        (3,),
        elements=st.floats(
            min_value=-10, max_value=10, allow_nan=False, allow_infinity=False
        ),
    ),
)
@pytest.mark.parametrize(
    "xp, device",
    testing_libs,
)
@pytest.mark.parametrize("image", [data.camera(), data.cells3d()[:, 0, :60, :60]])
def test_correctness_fourier_shift(
    xp,
    device,
    image,
    shift,
):
    input = np.fft.fftn(util.img_as_float(image))
    shift = shift[: image.ndim]

    input_xp = xp.asarray(input, device=device)
    shift_xp = xp.asarray(shift, device=device)

    output = fourier_shift(input_xp, shift_xp)
    output_scipy = fourier_shift_scipy(input, shift)

    assert_allclose(
        output,
        xp.asarray(output_scipy, device=device),
        rtol=1e-5,
    )


def test_broadcasting_fourier_shift():  # TODO
    pass



"""Some stuff are adapted from:
https://github.com/matterport/Mask_RCNN/blob/master/mrcnn/utils.py.
"""

import os
import pickle
from typing import Dict, List, Tuple

import numpy as np
from scipy.ndimage import center_of_mass
from skimage.exposure import equalize_adapthist
from skimage.filters import threshold_otsu
from tqdm import tqdm

from ..utils import center_to_corners, load_array, summary

# +------------------------------------------------------------------------------------------+ #
#                                           Volume                                           | #
# +------------------------------------------------------------------------------------------+ #


def compute_volume(boxes: np.ndarray) -> np.ndarray:
    """Compute the volume of each boxes in a given list.

    Args:
        boxes (np.ndarray): List of boxes. Array of shape (num_boxes, (x1, y1, z1, x2, y2, z2)).

    Returns:
        np.ndarray: (num_boxes, ): Volume of each boxe.
    """
    x1, y1, z1 = boxes[:, 0], boxes[:, 1], boxes[:, 2]
    x2, y2, z2 = boxes[:, 3], boxes[:, 4], boxes[:, 5]
    return (x2 - x1) * (y2 - y1) * (z2 - z1)


# +------------------------------------------------------------------------------------------+ #
#                                             IoU                                            | #
# +------------------------------------------------------------------------------------------+ #


def compute_iou_3d(
    box: np.ndarray, boxes: np.ndarray, box_volume: np.ndarray, boxes_volume: np.ndarray
) -> np.ndarray:
    """Calculates IoU of the given box with the array of the given boxes.

    Args:
        box (np.ndarray): 1D vector (x1, x1, z1, x2, y2, z2).
        boxes (np.ndarray): List of boxes. Array of shape (num_boxes, (x1, y1, z1, x2, y2, z2)).
        box_volume (np.ndarray): Volume of the given box: (x2 - x1) * (y2 - y1) * (z2 - z1)
        boxes_volume (np.ndarray): Volume of all the given boxes. 1D vector (num_boxes, ).

    Returns:
        np.ndarray: IoU between box and every other box in boxes. 1D vector (num_boxes, ).
                    Each element is a float32 in [0, 1].
    """
    x1, x2 = np.maximum(box[0], boxes[:, 0]), np.minimum(box[3], boxes[:, 3])
    y1, y2 = np.maximum(box[1], boxes[:, 1]), np.minimum(box[4], boxes[:, 4])
    z1, z2 = np.maximum(box[2], boxes[:, 2]), np.minimum(box[5], boxes[:, 5])
    intersection = (
        np.maximum(x2 - x1, 0) * np.maximum(y2 - y1, 0) * np.maximum(z2 - z1, 0)
    )
    union = box_volume + boxes_volume[:] - intersection[:]
    iou = intersection / union
    return iou


# +------------------------------------------------------------------------------------------+ #
#                                         Group Predictions                                  | #
# +------------------------------------------------------------------------------------------+ #


def group_predictions(boxes: np.ndarray) -> Tuple[List[np.ndarray]]:
    """Group predictions (boxes (+ scores)) that have a non zero IoU. This is a kind of
    connected components retrieval on boxes.
    Note that this is almost NMS itself with a threshold of 0 except that we don't take
    scores into account and remember the found group at each iteration.

    Args:
        boxes (np.ndarray): List of boxes. Array of shape (num_boxes, (x1, y1, z1, x2, y2, z2)).

    Returns:
        Tuple[List[np.ndarray]]: Grouped boxes and predictions. Each boxes group and scores group
                                 are of shape (k, (x1, y1, z1, x2, y2, z2)) and (k, ).
                                 The length of all groups sum up to num_boxes.
    """
    volumes = compute_volume(boxes)
    boxes_groups = []
    while len(boxes) > 0:
        reference_bbox = boxes[0]
        reference_volume = volumes[0]
        ious = compute_iou_3d(reference_bbox, boxes, reference_volume, volumes)
        group_indices = np.where(ious > 0)[0]
        boxes_groups.append(boxes[group_indices])
        boxes = np.delete(boxes, group_indices, axis=0)
        volumes = np.delete(volumes, group_indices)
    return boxes_groups


# +------------------------------------------------------------------------------------------+ #
#                                   Spatial Average Correction                               | #
# +------------------------------------------------------------------------------------------+ #


def spatial_average(grouped_boxes: List[np.ndarray]) -> np.ndarray:
    """Compute the mean boxes of each group in grouped_boxes.

    Args:
        grouped_boxes(List[np.ndarray]): List of group of boxes: (N, k, (x1, y1, z1, x2, y2, z2)).
                                         N is the number of groups, k_i is the number of boxes in
                                         the i-th group.

    Returns:
        np.ndarray: Mean boxe of each group: (N, (x1, y1, z1, x2, y2, z2)).
    """
    mean_boxes = np.ndarray((len(grouped_boxes), 6))
    for i, group in enumerate(grouped_boxes):
        x1, y1, z1 = group[:, 0], group[:, 1], group[:, 2]
        x2, y2, z2 = group[:, 3], group[:, 4], group[:, 5]
        mean_boxes[i] = np.array(
            (x1.mean(), y1.mean(), z1.mean(), x2.mean(), y2.mean(), z2.mean())
        )
    return mean_boxes


# +------------------------------------------------------------------------------------------+ #
#                                    Non Maximum Suppresion                                  | #
# +------------------------------------------------------------------------------------------+ #


def non_max_suppression_3d(
    boxes: np.ndarray, scores: np.ndarray, threshold: float
) -> np.ndarray:
    """Performs 3D non-maximum suppression and returns indices of kept boxes.

    Args:
        boxes (np.ndarray): List of boxes. Array of shape (num_boxes, (x1, y1, z1, x2, y2, z2)).
        scores (np.ndarray): 1-D array of box scores. (num_boxes, )
        threshold (float): IoU threshold to use for filtering. Float in [0, 1].

    Returns:
        np.ndarray: Indices of kept boxes. (k, ) where 0 < k < num_boxes.
    """
    # Compute box volumes
    volume = compute_volume(boxes)
    # Get indicies of boxes sorted by scores (highest first)
    ixs = scores.argsort()[::-1]
    pick = []
    while len(ixs) > 0:
        # Pick top box and add its index to the list
        i = ixs[0]
        pick.append(i)
        # Compute IoU of the picked box with the rest
        iou = compute_iou_3d(boxes[i], boxes[ixs[1:]], volume[i], volume[ixs[1:]])
        # Identify boxes with IoU over the threshold. This returns indices into ixs[1:],
        # so add 1 to get indices into ixs.
        remove_ixs = np.where(iou > threshold)[0] + 1
        # Remove indices of the picked and overlapped boxes.
        ixs = np.delete(ixs, remove_ixs)
        ixs = np.delete(ixs, 0)
    return np.array(pick, dtype=np.int32)


# +------------------------------------------------------------------------------------------+ #
#                                   Center of mass correction                                | #
# +------------------------------------------------------------------------------------------+ #


def get_center_of_mass(image: np.ndarray) -> np.ndarray:
    """Compute the center of mass of an image.
    Note that the center of mass is computed on a binary image obtained by Otsu thresholding.

    Args:
        image (np.ndarray): The 3D image array: (depth, height, width).

    Returns:
        np.ndarray: Coordinates of the center of mass: (center_z, center_y, center_x).
    """
    binary_image = image > threshold_otsu(image)
    return center_of_mass(binary_image)


def center_of_mass_correction(bbox: np.ndarray, image: np.ndarray) -> np.ndarray:
    """For one given bbox, compute the center of mass of its binary mask obtained by otsu
    thresholding, and create a new box centered on it.

    Args:
        bbox (np.ndarray): Array (x_min, y_min, z_min, x_max, y_max, z_max)
        image (np.ndarray): 3D Image into which boxes indicates objects: (depth, height, width).

    Returns:
        np.ndarray: New recentered bbox (x_min, y_min, z_min, x_max, y_max, z_max).
    """
    x_min, y_min, z_min, x_max, y_max, z_max = bbox.round().astype(int)
    crop = image[max(0, z_min) : z_max, max(0, y_min) : y_max, max(0, x_min) : x_max]
    center_relative = get_center_of_mass(crop)  # coordinates inside the crop
    center_absolute = np.array(center_relative) + np.array((z_min, y_min, x_min))
    depth, height, width = z_max - z_min, y_max - y_min, x_max - x_min
    corners = center_to_corners(center_absolute, (depth, height, width))
    return np.array(corners)


def iterative_center_of_mass_correction(
    bbox: np.ndarray,
    image: np.ndarray,
    min_distance: float = 0.1,
    n_iter_max: int = 1000,
) -> np.ndarray:
    d, h, w = image.shape
    # State initialization: two elements: center and distance
    x_min, y_min, z_min, x_max, y_max, z_max = bbox
    depth, height, width = z_max - z_min, y_max - y_min, x_max - x_min
    old_center = np.array((z_min + depth // 2, y_min + height // 2, x_min + width // 2))
    distance = min_distance + 1  # init so that first while test is True
    n_iter = 0
    while distance > min_distance and n_iter < n_iter_max:
        # compute new bouding box
        bbox = center_of_mass_correction(bbox, image)
        x_min, y_min, z_min, x_max, y_max, z_max = bbox
        depth, height, width = z_max - z_min, y_max - y_min, x_max - x_min
        # stop and revert previous step if new center is too close to the border
        if x_min < 0 or y_min < 0 or z_min < 0 or x_max > w or y_max > h or z_max > d:
            bbox = center_to_corners(old_center, (depth, height, width))
            break
        # update state
        new_center = np.array(
            (z_min + depth // 2, y_min + height // 2, x_min + width // 2)
        )
        distance = np.linalg.norm(new_center - old_center)
        old_center = new_center
        n_iter += 1
    return bbox


def center_of_mass_correction_on_all_bboxes(
    boxes: np.ndarray,
    image: np.ndarray,
    iterative: bool = True,
    min_distance: float = 0.1,
) -> np.ndarray:
    """For each given boxes, compute its center of mass (iteratively or not) and create a new
    box centered on it.

    Args:
        boxes (np.ndarray): List of boxes. Array of shape (num_boxes, (x1, y1, z1, x2, y2, z2)).
        image (np.ndarray): 3D Image into which boxes indicates objects: (depth, height, width).
        iterative (bool, optional): If True, the center of mass correction is applied iteratively
                                    until a stopping criterion is met. Default to True.
        min_distance (float, optional): If the center of mass correction is applied iteratively,
                                        it is so until the new center is at most at min_distance
                                        from the old one.

    Returns:
        np.ndarray: Corners array, of shape (num_boxes, (x1, y1, z1, x2, y2, z2)).
    """
    correction = (
        iterative_center_of_mass_correction if iterative else center_of_mass_correction
    )
    correction_kwargs = {"min_distance": min_distance} if iterative else {}
    corrected_boxes = []
    for bbox in boxes:
        corrected_bbox = correction(bbox, image, **correction_kwargs)
        if not np.isnan(corrected_bbox).any():
            corrected_boxes.append(corrected_bbox)
    return np.array(corrected_boxes)


def otsu_threshold(image: np.ndarray, boxes: np.ndarray, margin: float) -> np.ndarray:
    """Filtered boxes by computing a threshold via Otsu method on boxes intensities, after
    having corrected the image histogramm via CLAHE method.

    Args:
        image (np.ndarray): The image on wich CLAHE will be applied to compute boxes intensities.
        boxes (np.ndarray): List of boxes. Array of shape (num_boxes, (x1, y1, z1, x2, y2, z2)).
        margin (float): A box will be kept if its intensity is > margin * threshold.

    Returns:
        np.ndarray: Filtered corners array, of shape (new_num_boxes, (x1, y1, z1, x2, y2, z2)),
                    Where new_num_boxes <= num_boxes.
    """
    # 1. Correct image histogramm
    corrected_image = equalize_adapthist(image)
    # 2. Fetch boxes intensities
    intensities = []
    for corners in boxes:
        x_min, y_min, z_min, x_max, y_max, z_max = corners.astype(int)
        patch = corrected_image[z_min:z_max, y_min:y_max, x_min:x_max]
        intensities.append(patch.sum())
    intensities = np.array(intensities)
    # 3. Normalize intensities
    intensities = (intensities - intensities.min()) / (
        intensities.max() - intensities.min()
    )
    # 4. Compute Otsu threshold
    threshold = threshold_otsu(intensities)
    # 5. Filter boxes
    filtered_boxes = boxes[intensities > margin * threshold]
    return filtered_boxes


# +------------------------------------------------------------------------------------------+ #
#                            Refinements & Postprocessing Wrappers                           | #
# +------------------------------------------------------------------------------------------+ #


def refine(
    image: np.ndarray,
    boxes: np.ndarray,
    iterative: bool = True,
    min_distance: float = 0.1,
    last_step: str = "spatial_average",
    margin: float = 0.5,
) -> Tuple[np.ndarray]:
    """Being given an image and its predicted boxes, compute a set of refined boxes, that is:
        1. spatial average of each connected components
        2. iterative center of mass correction
        3.  * spatial average of each connected components
            OR
            * Otsu thresholding of boxes intensities computed on a image corrected by
           Contrast Limited Adaptive Histogram Equalization (CLAHE)
    Args:
        image (np.ndarray): 3D image being predicted: (depth, height, width).
        boxes (np.ndarray): Predicted bounding boxes: (num_boxes, (x1, y1, z1, x2, y2, z2)).
        iterative (bool, optional): If True, the center of mass correction is applied iteratively
                                    until a stopping criterion is met. Default to True.
        min_distance (float, optional): If the center of mass correction is applied iteratively,
                                        it is so until the new center is at most at min_distance
                                        from the old one.
        last_step (str): One of 'spatial_average' or 'otsu'. Determines which last postprocessing
                         step to applied after spatial average and center of mass correction.

    Returns:
        Tuple[np.ndarray]: Three corners array, each of shape (n, (x1, y1, z1, x2, y2, z2)),
                           after each postprocessing step.
    """
    # 1. Spatial Average of Connected components
    step1 = spatial_average(group_predictions(boxes))
    # 2. (Iterative) Center of mass Correction
    correction_args = [step1, image, iterative, min_distance]
    step2 = center_of_mass_correction_on_all_bboxes(*correction_args)
    # 3. Spatial Average of Connected components on corrected boxes OR Ostu after CLAHE
    if last_step == "otsu":
        step3 = otsu_threshold(image, step2, margin)
    else:
        step3 = spatial_average(group_predictions(step2))
    return step1, step2, step3


# +------------------------------------------------------------------------------------------+ #
# |                                MAIN POSTPROCESSING FUNCTION                              | #
# +------------------------------------------------------------------------------------------+ #


def postprocess_one_image(
    image: np.ndarray,
    boxes: np.ndarray,
    scores: np.ndarray,
    nms_threshold: float = 0.1,
    iterative: bool = True,
    last_step: str = "spatial_average",
    margin: float = 0.5,
) -> Dict:
    """Being given an image, bounding boxes, and scores, return a set of 'good' bounding boxes.
    The bounding boxes are filtered based on the following procedure:
        1. Non Maximum Suppresion
        2. Spatial Average of each Connected Components
        3. (Iterative) Center of mass correction, where the center of mass is computer on
           a binary mask obtained by otsu thresholding.
        4. Spatial Average of each resulting Connected Components

    Args:
        image (np.ndarray): 3D image being predicted: (depth, height, width). The image must
                              be padded the same way it was during inference, with respect to
                              given patch size and stride.
        boxes (np.ndarray): Predicted bounding boxes: (num_boxes, (x1, y1, z1, x2, y2, z2)).
        scores (np.ndarray): Confidence score of each bouding boxes: (num_boxes, ).
        nms_threshold (float, optional): IoU threshold above which it is considered that
                                         predictions overlap. Used in the NMS algorithm.
                                         Defaults to 0.1.
        iterative (bool, optional): If True, the center of mass correction is applied iteratively
                                    until a stopping criterion is met. Default to True.
        min_distance (float, optional): If the center of mass correction is applied iteratively,
                                        it is so until the new center is at most at min_distance
                                        from the old one.

    Returns:
        Dict: Corners array, of shape (num_boxes', (x1, y1, z1, x2, y2, z2)), with
              num_boxes' < num_boxes, after each postprocessing steps.
    """
    # NMS
    filtered_indices = non_max_suppression_3d(boxes, scores, threshold=nms_threshold)
    filtered_boxes = boxes[filtered_indices]
    # Refinement
    step1, step2, step3 = refine(
        image, filtered_boxes, iterative, nms_threshold, last_step, margin
    )
    results = {
        "raw": boxes,
        "nms": filtered_boxes,
        "first_spatial_average": step1,
        "center_of_mass_correction": step2,
        "last_step": step3,
    }
    return results


# +------------------------------------------------------------------------------------------+ #
# |                                       EXTERNAL CALL                                      | #
# +------------------------------------------------------------------------------------------+ #


def postprocess(
    predictions_path: str,
    rootdir: str,
    patch_size: Tuple[int],
    stride: Tuple[int],
    nms_threshold: float,
    iterative: bool,
    last_step: str,
    margin: float,
    extension: str,
    testset_size: int = None,
    image_name: str = None,
    output_dir: str = None,
    downscale: float = 1.0,
) -> Dict[str, Tuple[np.ndarray]]:
    """If image_name is None, postprocess whole folder."""
    patch_size = tuple(np.rint(np.array(patch_size) / downscale).astype(int))
    stride = (
        np.array(patch_size) // 4
        if stride is None
        else tuple(np.rint(np.array(stride) / downscale).astype(int))
    )
    summary_output = None
    if output_dir is not None:
        os.makedirs(output_dir, exist_ok=True)
        summary_output = os.path.join(output_dir, "postprocessing_config.txt")
    summary(locals(), title="POSTPROCESSING", output=summary_output)
    with open(predictions_path, "rb") as file:
        predictions = pickle.load(file)
    if image_name is not None:
        images_names = [image_name]
    else:
        images_names = sorted(
            list(filter(lambda x: x.endswith(extension), os.listdir(rootdir)))
        )
        images_names = images_names[:testset_size]
    postprocess_kwargs = {
        "nms_threshold": nms_threshold,
        "iterative": iterative,
        "last_step": last_step,
        "margin": margin,
    }
    results = {}
    for image_name in tqdm(images_names):
        centers, scores = predictions[image_name]
        corners = np.array(
            [center_to_corners(center, patch_size) for center in centers]
        )
        image = load_array(os.path.join(rootdir, image_name))
        results[image_name] = postprocess_one_image(
            image, corners, scores, **postprocess_kwargs
        )
    if output_dir is not None:
        with open(os.path.join(output_dir, "raw_predictions.pickle"), "wb") as file:
            pickle.dump(predictions, file)
        with open(os.path.join(output_dir, "predictions.pickle"), "wb") as file:
            pickle.dump(results, file)
    return results



import numpy as np
import pytest
import torch

from spfluo.utils.symmetrize_particle import symmetrize
from spfluo.utils.transform import get_transform_matrix
from spfluo.utils.volume import affine_transform, are_volumes_aligned


@pytest.mark.skip(reason="to fix")
def test_symmetrize_simple(generated_data_anisotropic):
    volumes, _, psf, groundtruth = generated_data_anisotropic
    vol = volumes[6]
    # translate vol
    vol = affine_transform(
        vol,
        np.linalg.inv(
            get_transform_matrix(
                vol.shape,
                np.array([0.0, 0.0, 0.0]),
                np.array([-7.0, 2.0, 5.0]),
                degrees=True,
            )
        ),
    )
    recon = symmetrize(
        torch.as_tensor(vol), (2.0, 5.0), 9, torch.as_tensor(psf), torch.as_tensor(1e-3)
    ).numpy()

    assert are_volumes_aligned(groundtruth, recon, atol=4)



import os
import random
from typing import Dict, Tuple

import imageio
import numpy as np
import torch


def seed_all(seed_numpy: bool = True) -> None:
    """For reproductibility.

    But, if we shuffle annotations between runs, we want each shuffling to be unique.
    Hence the option to choose if numpy must be seeded.
    """
    random.seed(0)
    torch.manual_seed(0)
    torch.cuda.manual_seed_all(0)
    if seed_numpy:
        np.random.seed(0)


def load_array(path: str) -> np.ndarray:
    """Takes a complete path to a file and return the corresponding numpy array.

    Args:
        path (str): Path to the file to read. It MUST contains the extension as this will
                    determines the way the numpy array is loaded from the file.

    Returns:
        np.ndarray: The array stored in the file described by 'path'.
    """
    extension = os.path.splitext(path)[-1]
    if extension == ".npz":
        return np.load(path)["image"]
    elif extension in [".tif", ".tiff"]:
        image = np.stack(imageio.mimread(path, memtest=False))
        # Some tiff images are heavily imbalanced: their data type is int16 but very few voxels
        # are actually > 255. If this is the case, the image in truncated and casted to uint8.
        if image.dtype == np.int16 and ((image > 255).sum() / image.size) < 1e-3:
            image[image > 255] = 255
            image = image.astype(np.uint8)
        else:
            image = image.astype(np.float32)
            image = 255 * (image - image.min()) / (image.max() - image.min())
            image = image.astype(np.uint8)
        return image
    error_msg = f"Found extension {extension}. Extension must be one of npz or tif."
    raise NotImplementedError(error_msg)


def load_annotations(csv_path: str) -> np.ndarray:
    """Csv containing coordinates of objects center.

    Args:
        csv_path (str): Path of the file to read.

    Returns:
        np.ndarray: Array into which each line is alike ('image name', particle_id, z, y, x).
    """
    with open(csv_path, "r") as f:
        lines = f.readlines()
    data = []
    for line in lines:
        line_data = line.split(",")
        line_data[2:] = list(map(float, line_data[2:]))
        data.append(line_data)
    return np.array(data, dtype=object)


def load_angles(csv_path: str) -> np.ndarray:
    """Csv containing angles of particles.

    Args:
        csv_path (str): Path of the file to read.

    Returns:
        np.ndarray: Array into which each line is alike ('image name', particle_id, z, y, x).
    """
    with open(csv_path, "r") as f:
        lines = f.readlines()
    data = []
    for line in lines:
        line_data = line.split(",")
        line_data[2:] = list(map(float, line_data[2:]))
        data.append(line_data)
    return np.array(data, dtype=object)


def center_to_corners(center: Tuple[int], size: Tuple[int]) -> Tuple[int]:
    depth, height, width = size
    center_z, center_y, center_x = center
    z_min = center_z - depth // 2
    y_min = center_y - height // 2
    x_min = center_x - width // 2
    z_max = z_min + depth
    y_max = y_min + height
    x_max = x_min + width
    return x_min, y_min, z_min, x_max, y_max, z_max


def summary(
    kwargs: Dict, title: str, output: str = None, return_table: bool = False
) -> None:
    # 1. Get key max length
    key_length = max([len(k) for k in kwargs.keys()])
    # 2. Get total length
    length = max([key_length + len(str(v)) for k, v in kwargs.items()]) + 4
    # 3. Define table delimiter and title
    hbar = "+" + length * "-" + "+" + "\n"
    pad_left = (length - len(title)) // 2
    pad_right = length - len(title) - pad_left
    title = "|" + pad_left * " " + title + pad_right * " " + "|" + "\n"
    # 4. Create table string
    table = "\n" + hbar + title + hbar
    # 5. Add lines to table
    for k, v in kwargs.items():
        line = k + (key_length - len(k)) * "." + ": " + str(v)
        line = "| " + line + (length - 1 - len(line)) * " " + "|" + "\n"
        table += line
    table += hbar
    # 6. Print table
    print(table)
    # 7. Save table to txt file (optional)
    if output is not None:
        with open(output, "w") as file:
            file.write(table)
    # 8. Return table (optional)
    if return_table:
        return table


def send_mail(subject: str, body: str) -> None:
    import smtplib
    import ssl

    port = 465  # SSL
    password = "hackitifyouwantidontcare"
    sender_email = "icuberemotedev@gmail.com"
    receiver_email = "vedrenneluc@gmail.com"
    message = f"""\
    {subject}

    {body}
    """
    context = ssl.create_default_context()
    with smtplib.SMTP_SSL("smtp.gmail.com", port, context=context) as server:
        server.login(sender_email, password)
        server.sendmail(sender_email, receiver_email, message)



import os

SEAFILE_URL = "https://seafile.unistra.fr"
REPO_ID = "c6458c20-0251-4c07-b296-e72240bd7317"
REPO_TOKEN = "6d6011eac0fe2f9e916d2c6a88f7e5fadad4ab10"
ARCHIVE_NAME = "data.zip"
UNISTRA_USERNAME = os.environ.get("UNISTRA_USERNAME")
UNISTRA_PASSWORD = os.environ.get("UNISTRA_PASSWORD")



import logging
from typing import TYPE_CHECKING, Optional

from spfluo.utils.array import array_namespace, get_device, to_device
from spfluo.utils.memory import split_batch
from spfluo.utils.volume import interpolate_to_size, pad

if TYPE_CHECKING:
    from spfluo.utils.array import Array, Device

logger = logging.getLogger("spfluo.refinement")


def deconv(
    image: "Array",
    psf: "Array",
    lambda_: "Array",
    device: Optional["Device"] = None,
    batch_size: Optional[int] = None,
):
    """Reconstruct a particule from volumes and their poses.
    M reconstructions can be done at once.

    Args:
        image (Array): stack of N 3D images of shape ((N), D, H, W)
        psf (Array) : 3D image of shape ((N), d, h, w)
        lambda_ (Array): regularization parameters of shape ((N),)
        device (Device): the device to do the computation on.
        batch_size (int or None) : if None, do all the computation at once

    Returns:
        deconv (Array):
            deconvolutions(s) of shape ((N), D, H, W) or ((N), D+1, H+1, W+1)
    """
    logger.info("Calling function deconv")

    xp = array_namespace(image, psf, lambda_)
    host_device = get_device(image)
    compute_device = device
    D, H, W = image.shape[-3:]
    d, h, w = psf.shape[-3:]

    batch = True
    if image.ndim == 3:
        image = image[None]
        batch = False
    if psf.ndim == 3:
        psf = psf[None]
    if lambda_.ndim == 0:
        lambda_ = lambda_[None]
    N = image.shape[0]
    assert psf.shape[0] == N
    assert lambda_.shape[0] == N

    logger.info(
        "Arguments:" f" {N} images of size {D}x{H}x{W}" f" PSF of size {d}x{h}x{w}"
    )

    # pad by 1 pixel to the right if even dimensions
    image = pad(image, ((0, 0), (0, (D + 1) % 2), (0, (H + 1) % 2), (0, (W + 1) % 2)))
    D_, H_, W_ = image.shape[-3:]
    logger.info(f"Reshaped volumes to odd size {D_}x{H_}x{W_}")

    psf = interpolate_to_size(psf, (D_, H_, W_), batch=True)

    num = xp.zeros((N, D_, H_, W_), dtype=xp.complex64, device=host_device)
    den = xp.zeros_like(num, dtype=xp.float32)

    dxyz = xp.zeros((3, 2, 2, 2), device=host_device, dtype=xp.complex64)
    dxyz[0, 0, 0, 0] = 1
    dxyz[0, 1, 0, 0] = -1
    dxyz[1, 0, 0, 0] = 1
    dxyz[1, 0, 1, 0] = -1
    dxyz[2, 0, 0, 0] = 1
    dxyz[2, 0, 0, 1] = -1

    dxyz_padded = pad(
        dxyz, ((0, 0), *[((x - 1) // 2, (x - 2) // 2) for x in (D_, H_, W_)])
    )
    DtD = xp.sum((xp.abs(xp.fft.fftn(dxyz_padded, axes=(1, 2, 3))) ** 2), axis=0)
    den += xp.asarray(lambda_[:, None, None, None] * DtD, dtype=xp.float32)
    del DtD

    # Move data to compute device
    for start, end in split_batch((N,), batch_size):
        images_minibatch = to_device(image[start:end, ...], compute_device)
        psf_minibatch = to_device(psf[start:end, ...], compute_device)

        y = xp.asarray(images_minibatch, dtype=xp.complex64)
        H_ = xp.asarray(psf_minibatch, dtype=xp.complex64)

        H_ = xp.fft.fftn(H_, axes=(-3, -2, -1))

        # Compute numerator
        y = xp.fft.fftn(xp.fft.fftshift(y, axes=(-3, -2, -1)), axes=(-3, -2, -1))
        y = xp.conj(H_) * y
        num[start:end, ...] += to_device(
            xp.sum(y, axis=0, dtype=xp.complex64), host_device
        )  # reduce N dims

        # Compute denominator
        den[start:end, ...] += to_device(
            xp.sum(xp.abs(H_) ** 2, axis=0, dtype=xp.float32), host_device
        )
        del H_

    num = xp.fft.ifftn(num / den, axes=(-3, -2, -1))
    deconv = xp.real(num)
    del num
    deconv = xp.where(deconv > 0, deconv, xp.asarray(0.0, device=host_device))

    deconv = interpolate_to_size(deconv, (D, H, W), batch=True)

    if not batch:
        deconv = deconv[0]

    return deconv



from functools import partial

import numpy as np
import pytest

from spfluo.data import generated_anisotropic
from spfluo.utils.separate import separate_centrioles, separate_centrioles_coords
from spfluo.utils.volume import assert_volumes_translated, resample


@pytest.fixture
def create_data():
    data = generated_anisotropic()
    im1, im2 = data["volumes"][[1, 2]]

    overlap = 15
    im = np.zeros((100 - overlap, 50, 50))
    im[:50] = im1
    im[50 - overlap :] += im2

    im = np.stack((im, np.random.randn(*im.shape)))
    im1 = np.stack((im1, im[1, :50]))
    im2 = np.stack((im2, im[1, 50 - overlap :]))
    good_channel = 0
    return im, (im1, im2), good_channel


@pytest.fixture
def create_data_monochannel(create_data):
    im, (im1, im2), _ = create_data
    return im[0], (im1[0], im2[0])


@pytest.fixture
def create_data_scaled(create_data_monochannel):
    im, (im1, im2) = create_data_monochannel
    scale = (3, 1, 1)
    im_resampled = resample(im, 1 / np.asarray(scale), order=3)
    im1_resampled = resample(im1, 1 / np.asarray(scale), order=3)
    im2_resampled = resample(im2, 1 / np.asarray(scale), order=3)
    return im_resampled, (im1_resampled, im2_resampled), scale


def common_im(im1, im2):
    common_shape = np.min((im1.shape, im2.shape), axis=0)
    return (
        im1[: common_shape[0], : common_shape[1], : common_shape[2]],
        im2[: common_shape[0], : common_shape[1], : common_shape[2]],
    )


def save_and_assert(save_result, im, im1, im2, im11, im22):
    save_result("original-im", im)
    save_result("im11", im11)
    save_result("im22", im22)

    ATOL = 1

    assert_volumes_translated_ = partial(assert_volumes_translated, atol=ATOL)
    same_order = assert_volumes_translated_(*common_im(im1, im11))
    if same_order:
        assert_volumes_translated_(*common_im(im2, im22))
    else:
        assert_volumes_translated_(*common_im(im1, im22))
        assert_volumes_translated_(*common_im(im2, im11))


def test_separate_simple(create_data_monochannel, save_result):
    im, (im1, im2) = create_data_monochannel
    im11, im22 = separate_centrioles(im, output_size=(50, 50, 50))

    save_and_assert(save_result, im, im1, im2, im11, im22)


def test_separate(create_data_monochannel, save_result):
    im, (im1, im2) = create_data_monochannel
    im11, im22 = separate_centrioles_coords(
        im, (50, 25, 25), (100, 50, 50), (50, 50, 50)
    )

    save_and_assert(save_result, im, im1, im2, im11, im22)


def test_separate_multichannel(create_data, save_result):
    im, (im1, im2), good_channel = create_data

    im11, im22 = separate_centrioles(im, (50, 50, 50), channel=good_channel)

    save_and_assert(
        save_result,
        im[good_channel],
        im1[good_channel],
        im2[good_channel],
        im11[good_channel],
        im22[good_channel],
    )


def test_separate_coords_multichannel(create_data, save_result):
    im, (im1, im2), good_channel = create_data

    im11, im22 = separate_centrioles_coords(
        im, (50, 25, 25), (100, 50, 50), (50, 50, 50), channel=good_channel
    )

    save_and_assert(
        save_result,
        im[good_channel],
        im1[good_channel],
        im2[good_channel],
        im11[good_channel],
        im22[good_channel],
    )


def test_separate_scaled(create_data_scaled, save_result):
    im, (im1, im2), scale = create_data_scaled

    im11, im22 = separate_centrioles_coords(
        im, (50, 25, 25), (100, 50, 50), (50, 50, 50), scale=scale
    )

    save_and_assert(save_result, im, im1, im2, im11, im22)



from typing import TYPE_CHECKING

import numpy as np
import pytest

import spfluo
from spfluo.ab_initio_reconstruction.api import AbInitioReconstruction
from spfluo.tests.helpers import assert_allclose, ids, testing_libs
from spfluo.utils.array import get_prefered_namespace_device, numpy
from spfluo.utils.transform import distance_family_poses
from spfluo.utils.volume import interpolate_to_size

if TYPE_CHECKING:
    from spfluo.utils.array import array_api_module

SEED = 123
GPU_LIBS = []
for x in ["torch-cuda", "cupy"]:
    try:
        GPU_LIBS.append(testing_libs[ids.index(x)][0])
    except ValueError:
        pass

params_minimal_run = {
    "N_iter_max": 1,
    "interp_order": 1,
    "N_axes": 2,
    "N_rot": 1,
}

params_long_run = {
    "N_iter_max": 4,
    "interp_order": 1,
}


def run(
    long: bool,
    xp: "array_api_module",
    gpu: bool,
    generated_data_anisotropic,
    tmpdir: str,
):
    volumes_arr, poses_arr, psf_array, groundtruth_array = generated_data_anisotropic
    np.random.seed(SEED)
    args = params_long_run if long else params_minimal_run
    ab_initio = AbInitioReconstruction(**args)
    psf_array = interpolate_to_size(psf_array, volumes_arr.shape[1:])
    _, device = get_prefered_namespace_device(xp=xp, gpu=gpu)
    ab_initio.fit(
        xp.asarray(volumes_arr, device=device),
        psf=xp.asarray(psf_array, device=device),
        output_dir=tmpdir,
    )

    return ab_initio, tmpdir


@pytest.fixture(scope="module")
def gpu_run(request, generated_data_anisotropic, tmp_path_factory):
    xp, long = request.param
    run_name = "long-run" if long else "minimal-run"
    run_name += f"_{xp.__name__}"
    tmpdir = tmp_path_factory.mktemp(run_name)
    return run(long, xp, True, generated_data_anisotropic, tmpdir)


@pytest.fixture(scope="module")
def numpy_run(request, generated_data_anisotropic, tmp_path_factory):
    long = request.param
    run_name = "long-run_numpy" if long else "minimal-run_numpy"
    tmpdir = tmp_path_factory.mktemp(run_name)
    return run(long, numpy, False, generated_data_anisotropic, tmpdir)


@pytest.mark.parametrize("numpy_run", [False], indirect=True)
def test_files_exist(numpy_run):
    ab_initio, tmpdir = numpy_run
    files = [
        "energies.csv",
        "params_learning_alg.json",
        "final_recons.ome.tiff",
        "poses.csv",
    ]
    for f in files:
        assert (tmpdir / f).exists()
    assert (tmpdir / "intermediar_results").exists()
    assert (tmpdir / "distributions_angles").exists()
    assert (tmpdir / "energies").exists()
    for i in range(1, ab_initio._num_iter):
        assert (
            tmpdir / "intermediar_results" / f"estimated_poses_epoch_{i}.csv"
        ).exists()
        assert (tmpdir / "intermediar_results" / f"recons_epoch_{i}.tif").exists()
        assert (tmpdir / "distributions_angles" / f"iter={i:04}_axes.npy").exists()
        assert (tmpdir / "distributions_angles" / f"iter={i:04}_rot.npy").exists()
        assert (tmpdir / "energies" / "energies_each_view_iter=0001.npy").exists()
    assert ab_initio._num_iter == len(ab_initio._energies)


@pytest.mark.parametrize(
    "gpu_run, numpy_run", [((lib, False), False) for lib in GPU_LIBS], indirect=True
)
def test_same_results_gpu(gpu_run, numpy_run):
    ab_initio_gpu, _ = gpu_run
    ab_initio_numpy, _ = numpy_run
    assert_allclose(ab_initio_numpy._energies, ab_initio_gpu._energies, rtol=0.002)


@pytest.mark.slow
@pytest.mark.skipif(
    not (spfluo.has_torch() and spfluo.has_torch_cuda()),
    reason="Too long to test if CUDA is not available",
)
@pytest.mark.parametrize("gpu_run", [(GPU_LIBS[0], True)], indirect=True)
def test_energy_threshold(gpu_run):
    ab_initio, _ = gpu_run
    assert ab_initio._energies[-1] < 210


@pytest.mark.skip(reason="not finished")
@pytest.mark.parametrize("gpu_run", [(GPU_LIBS[0], True)], indirect=True)
def test_poses_aligned(gpu_run):
    ab_initio, folder = gpu_run
    distance_family_poses



import numpy as np

from .volume_representation.gaussian_mixture_representation.GMM_grid_evaluation import (
    make_grid,
    nd_gaussian,
)


class ParametersMainAlg:
    def __init__(
        self,
        M_axes=360**2,
        M_rot=360,
        dec_prop=1.2,
        init_unif_prop=(1, 1),
        coeff_kernel_axes=50.0,
        coeff_kernel_rot=5.0,
        eps=0,
        lr=0.1,
        N_axes=25,
        N_rot=20,
        prop_min=0,
        interp_order=3,
        N_iter_max=20,
        gaussian_kernel=True,
        N_iter_with_unif_distr=None,
        epochs_of_suppression=None,
        proportion_of_views_suppressed=None,
        convention="XZX",
        dtype="float32",
        reg_coeff=0,
        beta_sampling=0,
        epoch_length=None,
        batch_size=1,
        beta_grad=0,
        random_sampling=False,
    ):
        self.params = locals()
        self.M_axes = M_axes
        self.M_rot = M_rot
        self.dec_prop = dec_prop
        self.coeff_kernel_axes = coeff_kernel_axes
        self.coeff_kernel_rot = coeff_kernel_rot
        self.eps = eps
        self.lr = lr
        self.N_axes = N_axes
        self.N_rot = N_rot
        self.prop_min = prop_min
        self.interp_order = interp_order
        self.gaussian_kernel = gaussian_kernel
        self.N_iter_with_unif_distr = N_iter_with_unif_distr
        self.epochs_of_suppression = epochs_of_suppression
        self.proportions_of_views_suppressed = proportion_of_views_suppressed
        self.N_iter_max = N_iter_max
        self.convention = convention
        self.init_unif_prop = init_unif_prop
        self.dtype = dtype
        self.reg_coeff = reg_coeff
        self.beta_sampling = beta_sampling
        self.epoch_length = epoch_length
        self.batch_size = batch_size
        self.beta_grad = beta_grad
        self.random_sampling = random_sampling


class ParametersDataGeneration:
    def __init__(
        self,
        nb_dim=3,
        nb_views=10,
        sig_z=5,
        sig_xy=1,
        snr=100,
        order=3,
        sigma_trans_ker=0,
        size=50,
        psf=None,
        projection=False,
        rot_vecs=None,
        convention="zxz",
        partial_labelling=False,
        **partial_labelling_args,
    ):
        self.params = locals()
        self.nb_views = nb_views
        self.sig_z = sig_z
        self.sig_xy = sig_xy
        self.snr = snr
        self.sigma_trans_ker = sigma_trans_ker
        self.size = size
        self.grid_step = 2 / (size - 1)
        self.psf = psf
        self.partial_labelling = partial_labelling
        self.partial_labelling_args = partial_labelling_args
        self.projection = projection
        self.rot_vecs = rot_vecs
        self.convention = convention
        self.nb_dim = nb_dim
        self.order = order

    def get_cov_psf(self):
        cov_PSF = self.grid_step**2 * np.eye(self.nb_dim)
        cov_PSF[0, 0] *= self.sig_z**2
        for i in range(1, self.nb_dim):
            cov_PSF[i, i] *= self.sig_xy**2
        return cov_PSF

    def get_psf(self):
        if self.psf is None:
            cov_psf = self.get_cov_psf()
            grid = make_grid(self.size, self.nb_dim)
            psf = nd_gaussian(grid, np.zeros(self.nb_dim), cov_psf, self.nb_dim)
            psf /= np.sum(psf)
            return psf
        else:
            return self.psf / np.sum(self.psf)


class ParametersGMM:
    def __init__(
        self,
        nb_gaussians_init=10,
        nb_gaussians_ratio=4,
        sigma_init=0.2,
        sigma_ratio=1.4,
        nb_steps=4,
        threshold_gaussians=0.01,
        unif_prop_mins=[0.5, 0.25, 0.125, 0],
        init_with_views=True,
    ):
        self.nb_gaussians_init = nb_gaussians_init
        self.nb_gaussians_ratio = nb_gaussians_ratio
        self.sigma_init = sigma_init
        self.sigma_ratio = sigma_ratio
        self.nb_steps = nb_steps
        self.threshold_gaussians = threshold_gaussians
        self.unif_prop_mins = unif_prop_mins
        self.init_with_views = init_with_views



from functools import partial
from typing import TYPE_CHECKING, Tuple

from scipy.spatial.transform import Rotation

from .array import array_namespace, get_device, median, numpy_only_compatibility

if TYPE_CHECKING:
    from spfluo.utils.array import Array


@numpy_only_compatibility
def euler_to_matrix(convention: str, euler_angles: "Array", degrees=False) -> "Array":
    """
    Params:
        convention: str
        euler_angles: Array of shape (3,) or (N, 3)
        degrees: bool
    """
    return Rotation.from_euler(convention, euler_angles, degrees=degrees).as_matrix()


@numpy_only_compatibility
def matrix_to_euler(convention: str, matrix: "Array", degrees: bool = False):
    """
    Params:
        convention: str
        euler_angles: Array of shape (3,) or (N, 3)
        degrees: bool
    """
    return Rotation.from_matrix(matrix).as_euler(convention, degrees=degrees)


def compose_poses(pose1: "Array", pose2: "Array", convention="XZX"):
    """Returns a pose: pose = pose2 ○ pose1
    Params:
        pose1, pose2: Array of shape (3,) or (N, 3)
        convention: str
    """
    xp = array_namespace(pose1, pose2)
    to_matrix = partial(euler_to_matrix, convention, degrees=True)
    to_euler = partial(matrix_to_euler, convention, degrees=True)
    new_pose = xp.zeros_like(pose1) + xp.zeros_like(pose2)
    new_pose[..., :3] = to_euler(to_matrix(pose2[..., :3]) @ to_matrix(pose1[..., :3]))
    new_pose[..., 3:] = (
        pose2[..., 3:] + (to_matrix(pose2[..., :3]) @ pose1[..., 3:, None])[..., 0]
    )
    return new_pose


def invert_pose(pose: "Array", convention="XZX"):
    xp = array_namespace(pose)
    inv_mat = xp.linalg.matrix_transpose(
        euler_to_matrix(convention, pose[..., :3], degrees=True)
    )
    t = pose[..., 3:]
    new_pose = xp.zeros_like(pose)
    new_pose[..., :3] = matrix_to_euler(convention, inv_mat, degrees=True)
    new_pose[..., 3:] = -(xp.astype(inv_mat, t.dtype) @ t[..., None])[..., 0]
    return new_pose


def get_transform_matrix_around_center(
    shape: Tuple[int, ...], rotation_matrix: "Array"
):
    """
    Params:
        - shape
        - rotation_matrix: Array of shape (..., ndim, ndim)
    Returns:
        Array of shape (..., ndim+1, ndim+1)
    """
    xp = array_namespace(rotation_matrix)
    array_kwargs = {
        "dtype": rotation_matrix.dtype,
        "device": get_device(rotation_matrix),
    }
    assert rotation_matrix.ndim >= 2
    ndim = rotation_matrix.shape[-1]
    assert ndim == rotation_matrix.shape[-2]
    assert ndim == len(shape)
    center = (xp.asarray(shape, **array_kwargs) - 1) / 2
    ndim = rotation_matrix.shape[-1]
    H_rot = xp.zeros(
        (rotation_matrix.shape[:-2]) + (ndim + 1, ndim + 1), **array_kwargs
    )
    H_rot[..., ndim, ndim] = 1.0
    H_center = xp.asarray(H_rot, copy=True)
    H_center[..., :ndim, ndim] = -center  # 1. translation to (0,0,0)
    for i in range(ndim):
        H_center[..., i, i] = 1.0  # diag to 1
    H_rot[..., :ndim, :ndim] = rotation_matrix  # 2. rotation
    H_rot[..., :ndim, ndim] = center  # 3. translation to center of image.

    #   3-2 <- 1
    H = H_rot @ H_center
    return H


def get_transform_matrix(
    shape: Tuple[int, int, int],
    euler_angles: "Array",
    translation: "Array",
    convention: str = "XZX",
    degrees: bool = False,
    dtype: str = "float64",
):
    """
    Returns the transformation matrix in pixel coordinates.
    The transformation is the composition of a rotation (defined by 3 euler angles),
    and a translation (defined by a translation vector).
    The rotation is made around the center of the volume.
    Params:
        shape: Tuple[int, int, int]
            shape of the image to be transformed. D, H, W
        euler_angles: np.ndarray of shape ((N), 3)
            𝛗, 𝛉, 𝛙. See convention to see how they are used.
        translation: np.ndarray of shape ((N), 3)
        convention: str
            Euler angles convention in scipy terms.
            See `scipy.spatial.transform.Rotation`.
            Default to 'XZX'

                   a-------------b       numpy coordinates of points:
                  /             /|        - a = (0, 0, 0)
                 /             / |        - b = (0, 0, W-1)
                c-------------+  |        - c = (0, H-1, 0)
                |             |  |        - d = (D-1, H-1, 0)
                |             |  |        - e = (D-1, H-1, W-1)
                |             |  +
                |             X Y
                |             ↑↗
                d-----------Z←e   <-- reference frame used for rotations.
                                      The center of the rotation is at (D/2, H/2, W/2).

            If the convention 'XZX' is used:
                - first, rotate by 𝛗 around the X-axis. The XYZ frame is also rotated!
                - then, rotate by 𝛉 around the Z-axis.
                - finally, rotate by 𝛙 around the X-axis.

        degrees: bool
            Are the euler angles in degrees?
    Returns:
        np.ndarray of shape ((N), 4, 4)
        An (or N) affine tranformation(s) in homogeneous coordinates.
    """
    xp = array_namespace(euler_angles, translation)
    dtype = getattr(xp, dtype)
    assert xp.isdtype(dtype, "real floating")
    rot = euler_to_matrix(convention, euler_angles, degrees=degrees)
    H_rot = get_transform_matrix_around_center(shape, rot)
    H_rot[..., :3, 3] += translation  # adds the translation

    return xp.asarray(H_rot, dtype=dtype)


def get_transform_matrix_from_pose(
    shape: Tuple[int, int, int],
    pose: "Array",
    *,
    convention: str = "XZX",
):
    return get_transform_matrix(
        shape, pose[..., :3], pose[..., 3:], convention=convention, degrees=True
    )


def symmetrize_angles(
    euler_angles: "Array", symmetry: int, convention: str = "XZX", degrees: bool = False
) -> "Array":
    """
    axis of symmetry is around the X-axis (see get_transform_matrix)
    the euler_angles are in the XZX convention, so that the third angle is purely the
    angle of symmetry.
    Params:
        - euler_angles: Array of shape (..., 3)
        - symmetry: int
            degree of the symmetry
    Returns:
        euler_angles_sym: Array of shape (..., symmetry, 3)
    """
    assert convention == "XZX"
    xp = array_namespace(euler_angles)
    full_range = 360 if degrees else 2 * xp.pi
    symmetry_offset = xp.asarray(
        [i * full_range / symmetry for i in range(symmetry)],
        device=get_device(euler_angles),
        dtype=euler_angles.dtype,
    )  # shape (symmetry,)
    euler_angles_sym = xp.asarray(euler_angles, copy=True)
    euler_angles_sym = xp.stack(
        (euler_angles,) * symmetry, axis=-2
    )  # shape (..., symmetry, 3)
    euler_angles_sym[..., 2] += symmetry_offset
    return euler_angles_sym


def symmetrize_poses(poses: "Array", symmetry: int, convention: str = "XZX") -> "Array":
    """
    Params:
        - poses: Array of shape (..., 6)
        - symmetry: int
            degree of the symmetry
    Returns:
        - poses_sym: shape (..., k, 6)
    """
    assert convention == "XZX"
    euler_angles_sym = symmetrize_angles(
        poses[..., :3], symmetry=symmetry, degrees=True
    )  # shape (..., k, 3)
    xp = array_namespace(euler_angles_sym)
    poses_sym = xp.concat(
        (euler_angles_sym, xp.zeros_like(euler_angles_sym)), axis=-1
    )  # shape (..., k, 6)
    poses_sym[..., 3:] = poses[..., None, 3:]
    return poses_sym


def distance_poses(
    p1: "Array", p2: "Array", convention: str = "XZX", symmetry: int = 1
):
    """Compute the rotation distance and the euclidean distance between p1 and p2.
    Parameters:
        p1, p2 : arrays of shape (..., 6). Must be broadcastable.
            Represents poses (theta,psi,gamma,tz,ty,tx).
    Returns:
        distances : Tuple[Array, Array] of shape broadcasted dims.
    """
    # Rotation distance
    xp = array_namespace(p1, p2)
    dtype, device = p1.dtype, get_device(p1)
    euler1, euler2 = xp.asarray(p1[..., :3]), xp.asarray(p2[..., :3])
    rot_mat1 = xp.reshape(
        euler_to_matrix(convention, xp.reshape(euler1, (-1, 3)), degrees=True),
        euler1.shape[:-1] + (3, 3),
    )  # shape (..., 3, 3)
    rot_mat2 = xp.reshape(
        euler_to_matrix(convention, xp.reshape(euler2, (-1, 3)), degrees=True),
        euler2.shape[:-1] + (3, 3),
    )  # shape (..., 3, 3)
    sym_euler = xp.zeros((symmetry, 3), dtype=dtype, device=device)
    sym_euler[:, 2] = (
        -2 * xp.arange(symmetry, dtype=dtype, device=device) * xp.pi / symmetry
    )
    sym_matrices = euler_to_matrix(
        convention, sym_euler, degrees=False
    )  # shape (s, 3, 3)
    R = (rot_mat1[..., None, :, :] @ sym_matrices) @ xp.linalg.matrix_transpose(
        rot_mat2[..., None, :, :]
    )  # shape (..., s, 3, 3)
    traces = xp.sum(R[..., [0, 1, 2], [0, 1, 2]], axis=-1)  # shape (..., s)
    traces[traces > 3.0] = 3.0
    traces[traces < -1.0] = -1.0
    angles = xp.acos((traces - 1) / 2) * 180 / xp.pi  # shape (..., s)
    best_s = xp.argmin(xp.abs(xp.mean(angles, axis=tuple(range(angles.ndim - 1)))))
    rot_distance = angles[..., best_s]  # shape (...)

    # Euclidian distance
    t1, t2 = p1[..., 3:], p2[..., 3:]
    trans_distance = xp.sum(((t1 - t2) ** 2), axis=-1) ** 0.5

    return rot_distance, trans_distance


def distance_family_poses(
    guessed_poses: "Array",
    gt_poses: "Array",
    convention: str = "XZX",
    symmetry: int = 1,
):
    """Compute the rotation distance and the euclidean distance between guessed_poses
    and gt_poses.
    Account for an eventual offset in the guessed_poses.
        E.g. if the guessed poses are the same as the gt poses but rotated by R0, then
        the rotation distance will be 0.
    If symmetry is greater than 1, account also for symmetries.
        The symmetry must be around the first axis.

    Parameters:
        guessed_poses, gt_poses : arrays of shape (N, 6).
            Represents poses (theta,psi,gamma,tz,ty,tx).
        convention : string
        symmetry : int
    Returns:
        rotation distance, translation distance : Tuple[Array, Array] of shape (N,)
    """
    xp = array_namespace(guessed_poses, gt_poses)

    # Rotation distances
    # 1. convert euler angles to matrices
    euler1, euler2 = xp.asarray(guessed_poses[:, :3]), xp.asarray(gt_poses[:, :3])
    N, _ = euler1.shape
    guessed_rot_mat = euler_to_matrix(convention, euler1, degrees=True)
    gt_rot_mat = euler_to_matrix(convention, euler2, degrees=True)
    sym_euler = xp.zeros(
        (symmetry, 3), dtype=guessed_rot_mat.dtype, device=get_device(guessed_rot_mat)
    )
    sym_euler[:, 0] = (
        -2
        * xp.arange(
            symmetry, dtype=guessed_rot_mat.dtype, device=get_device(guessed_rot_mat)
        )
        * xp.pi
        / symmetry
    )
    sym_matrices = euler_to_matrix(convention, sym_euler, degrees=False)

    basis_change = (
        xp.linalg.matrix_transpose(guessed_rot_mat[None, :] @ sym_matrices[:, None])
        @ gt_rot_mat[None, :]
    )  # shape (s, N, 3, 3)
    assert basis_change.shape == (symmetry, N, 3, 3)
    diff = (
        guessed_rot_mat[None, :, None]  # shape (1, N, 1, 3, 3)
        @ basis_change[:, None]  # shape (s, 1, N, 3, 3)
        @ xp.linalg.matrix_transpose(gt_rot_mat[None, :, None])  # shape (1, N, 1, 3, 3)
    )
    # shape (s, N, N, 3, 3)
    traces = xp.sum(diff[:, :, :, [0, 1, 2], [0, 1, 2]], axis=-1)  # shape (s, N, N)
    traces[traces > 3.0] = 3.0
    angles = xp.acos((traces - 1) / 2)
    angles = xp.min(angles, axis=0)  # shape (s, N, N)
    median_angles = median(angles, axis=-1, xp=xp)
    median_angles = median_angles * 180 / xp.pi

    # Translation distances, simple L2 norm
    t1, t2 = xp.asarray(guessed_poses[..., 3:]), xp.asarray(gt_poses[..., 3:])
    trans_distances = xp.sum(((t1 - t2) ** 2), axis=-1) ** 0.5

    return median_angles, trans_distances



from .augmentations.transforms import (
    CenterCrop,
    Contiguous,
    CropNonEmptyMaskIfExists,
    ElasticTransform,
    ElasticTransformPseudo2D,
    Flip,
    Float,
    GaussianNoise,
    Normalize,
    PadIfNeeded,
    RandomCrop,
    RandomGamma,
    RandomResizedCrop,
    RandomRotate90,
    RandomScale,
    RandomScale2,
    RemoveEmptyBorder,
    Resize,
    ResizedCropNonEmptyMaskIfExists,
    Rotate,
    RotatePseudo2D,
    Transpose,
)
from .core.composition import Compose

__all__ = [
    Float,
    Contiguous,
    PadIfNeeded,
    GaussianNoise,
    Resize,
    RandomScale,
    RandomScale2,
    RotatePseudo2D,
    RandomRotate90,
    Flip,
    Normalize,
    Transpose,
    CenterCrop,
    RandomResizedCrop,
    RandomCrop,
    CropNonEmptyMaskIfExists,
    ResizedCropNonEmptyMaskIfExists,
    RandomGamma,
    ElasticTransformPseudo2D,
    ElasticTransform,
    Rotate,
    RemoveEmptyBorder,
    Compose,
]



import copy
import os

import numpy as np
import pandas as pd
import torch
import torch.multiprocessing
from torch.optim import Adam
from torch.utils.data import default_collate
from torch.utils.tensorboard import SummaryWriter
from torchvision.utils import make_grid
from tqdm.auto import tqdm

from .data import (
    Generator_variableSize,
    GeneratorDataset,
    HardCopyDataset,
    MultiEpochsDataLoader,
    infos_to_dataframe,
    unbatch_dict,
)
from .losses import iaawnLossDenis
from .unet_3d import UNet3D

if __name__ == "__main__":
    torch.multiprocessing.set_sharing_strategy("file_system")
    torch.backends.cudnn.benchmark = True
    num_epochs = 20
    lr = 1e-5
    alpha = 1
    beta = 1e-2
    n_samples_per_epoch = 200
    batch_size = 4
    n_samples_validation_sets = 50
    n_samples_test_sets = 1000

    n_training = 60
    exp_name = "exptest3-reg4"

    # Fix randomness
    torch.random.manual_seed(0)

    print("Init train loader")
    train_loader_a = MultiEpochsDataLoader(
        GeneratorDataset(Generator_variableSize(600), n_samples_per_epoch),
        batch_size=batch_size,
        num_workers=32,
        pin_memory=True,
        persistent_workers=True,
    )
    train_loader_a2 = MultiEpochsDataLoader(
        GeneratorDataset(Generator_variableSize(100, 1000), n_samples_per_epoch),
        batch_size=batch_size,
        num_workers=32,
        pin_memory=True,
        persistent_workers=True,
    )
    train_loader_b = MultiEpochsDataLoader(
        GeneratorDataset(Generator_variableSize(150), n_samples_per_epoch),
        batch_size=batch_size,
        num_workers=32,
        pin_memory=True,
        persistent_workers=True,
    )

    print("Init valid loaders")
    print("A...")
    valid_loader_a = torch.utils.data.DataLoader(
        HardCopyDataset(
            GeneratorDataset(Generator_variableSize(600), n_samples_validation_sets),
            num_workers=32,
        ),
        batch_size=8,
        pin_memory=True,
    )
    print("B...")
    valid_loader_b = torch.utils.data.DataLoader(
        HardCopyDataset(
            GeneratorDataset(Generator_variableSize(150), n_samples_validation_sets),
            num_workers=32,
        ),
        batch_size=8,
        pin_memory=True,
    )
    print("A extended...")
    valid_loader_a2 = torch.utils.data.DataLoader(
        HardCopyDataset(
            GeneratorDataset(
                Generator_variableSize(100, 1000), n_samples_validation_sets
            ),
            num_workers=32,
        ),
        batch_size=8,
        pin_memory=True,
    )

    print("Init testing loader...")
    testing_loader = torch.utils.data.DataLoader(
        HardCopyDataset(
            GeneratorDataset(Generator_variableSize(100, 1000), n_samples_test_sets),
            num_workers=32,
        ),
        batch_size=8,
        pin_memory=True,
        num_workers=8,
    )

    loss = iaawnLossDenis(normalization="sigmoid", alpha=alpha, beta=beta)
    metric = iaawnLossDenis(normalization="sigmoid", alpha=1, beta=1e-2)

    for name_train, train_loader, valid_loader_early_stopping in zip(
        ["A", "C", "B"],
        [train_loader_a, train_loader_a2, train_loader_b],
        [valid_loader_a, valid_loader_a2, valid_loader_b],
    ):
        # for name_train, train_loader in zip(["A"], [train_loader_a]):
        print(f"Start {name_train} training")
        for i in range(n_training):
            print("Init Unet...")
            unet = UNet3D(1, 1).to("cuda")
            print("Init optimizer...")
            optimizer = Adam(unet.parameters(), lr=lr)

            # Logs
            writer = SummaryWriter(comment=f"_{exp_name}-{name_train}-{i}")

            metrics = {"val-A": [], "val-B": [], "val-C": [], "train": []}
            best_loss = np.inf
            for j in range(num_epochs):
                # Validate
                print("Validating...")
                unet.eval()

                for name, valid_loader in zip(
                    ["val-A", "val-C", "val-B"],
                    [valid_loader_a, valid_loader_b, valid_loader_a2],
                ):
                    # for name, valid_ds in zip(["A"], [valid_ds_a]):
                    losses = []
                    images = []
                    for ims, masks, _ in valid_loader:
                        ims, masks = ims.to("cuda"), masks.to("cuda")
                        with torch.no_grad():
                            masks_out = unet(ims)[:, 0]
                            for m1, m2 in zip(masks_out, masks):
                                losses.append(metric(m1, m2).item())
                        for im, mask, mask_out in zip(ims[:, 0], masks, masks_out):
                            half_depth = im.size(1) // 2
                            im = im / im.max()
                            raw_im = im[:, half_depth].cpu().clone().expand(3, -1, -1)
                            pred_mask_im = raw_im.clone()
                            gt_mask_im = raw_im.clone()

                            # print(mask_out.dtype, mask.dtype)
                            pred_mask_im[:, mask_out[half_depth] > 0.5] = torch.tensor(
                                [0, 0, 1]
                            ).float()[:, None]
                            pred_mask_im[
                                :,
                                torch.logical_and(
                                    mask_out[half_depth] > 0.5, mask[half_depth] == 2
                                ),
                            ] = torch.tensor([0.8, 0, 0.8]).float()[:, None]
                            pred_mask_im[
                                :,
                                torch.logical_and(
                                    mask_out[half_depth] > 0.5, mask[half_depth] == 1
                                ),
                            ] = torch.tensor([0, 1, 1]).float()[:, None]
                            gt_mask_im[:, mask[half_depth] == 1] = torch.tensor(
                                [0, 1, 0]
                            ).float()[:, None]
                            gt_mask_im[:, mask[half_depth] == 2] = torch.tensor(
                                [1, 0, 0]
                            ).float()[:, None]

                            grid = make_grid([raw_im, pred_mask_im, gt_mask_im])
                            images.append(grid)

                    dice_mean = np.mean(losses)
                    metrics[name].append(dice_mean)
                    dice_std = np.std(losses)
                    writer.add_scalar(name, dice_mean, i)
                    writer.add_images(
                        f"validation {name} images", torch.stack(images), i
                    )
                print(
                    f"Validation epoch {j+1}, ",
                    f"validation set Loss mean : {dice_mean:.4f}, ",
                    f"Loss std : {dice_std:.4f}",
                )

                # Train
                print("Training...")
                unet.train()
                losses = []
                for im, mask, _ in tqdm(
                    train_loader,
                    total=n_samples_per_epoch // batch_size,
                    desc=f"Train epoch {i+1}",
                ):
                    optimizer.zero_grad()
                    im, mask = im.to("cuda"), mask.to("cuda")
                    mask_out = unet(im)[:, 0]
                    loss = loss(mask_out, mask)
                    loss.backward()
                    optimizer.step()

                    losses.append(copy.deepcopy(loss.detach().item()))
                    del im, mask, _, loss, mask_out
                dice_mean = np.mean(losses)
                writer.add_scalar("train loss", dice_mean, i)
                metrics["train"].append(dice_mean)

                # Evaluate
                print("Evaluate...")
                loss_ = 0
                for ims, masks, _ in valid_loader_early_stopping:
                    ims, masks = ims.to("cuda"), masks.to("cuda")
                    with torch.no_grad():
                        masks_out = unet(ims)[:, 0]
                        for m1, m2 in zip(masks_out, masks):
                            loss_ += metric(m1, m2).item()
                # early stopping, shouldn't be necessary
                if loss_ < best_loss:
                    best_loss = loss_
                    torch.save(
                        unet.state_dict(), os.path.join(writer.log_dir, "unet.pth")
                    )

            # Save training metrics
            pd.DataFrame(metrics).to_csv(
                os.path.join(writer.log_dir, "train_metrics.csv")
            )

            # Testing the model
            print("Testing...")
            unet = UNet3D(1, 1).to("cuda")
            unet.load_state_dict(torch.load(os.path.join(writer.log_dir, "unet.pth")))
            metadatas = []
            losses = []
            for ims, masks, infos in testing_loader:
                ims, masks = ims.to("cuda"), masks.to("cuda")
                with torch.no_grad():
                    masks_out = unet(ims)[:, 0]
                    for m1, m2, info in zip(masks_out, masks, unbatch_dict(infos)):
                        info["loss"] = metric(m1, m2).item()
                        metadatas.append(info)

            metadatas = default_collate(metadatas)
            infos_to_dataframe(metadatas).to_csv(
                os.path.join(writer.log_dir, "test_metrics.csv")
            )



from __future__ import annotations

import functools
from typing import TYPE_CHECKING, Any, overload

import numpy
from array_api_compat import (
    array_namespace as _array_namespace,
)
from array_api_compat import (
    device as get_device,
)
from array_api_compat import (
    is_array_api_obj,
    is_cupy_array,
    is_numpy_array,
    is_torch_array,
    to_device,
)

import spfluo

API_VERSION = "2022.12"

if int(numpy.__version__[0]) < 2:
    from array_api_compat import numpy


def array_namespace(*xs):
    """
    Get the array API compatible namespace for the arrays `xs`.

    `xs` should contain one or more arrays.

    Typical usage is

        def your_function(x, y):
            xp = array_namespace(x, y)
            # Now use xp as the array library namespace
            return xp.mean(x, axis=0) + 2*xp.std(y, axis=0)
    """
    try:
        xp = _array_namespace(*xs, api_version=API_VERSION, use_compat=True)
    except ValueError:
        xp = _array_namespace(*xs, api_version=API_VERSION, use_compat=False)
    return xp


def get_torch():
    from array_api_compat import torch

    return torch


def get_cupy():
    from array_api_compat import cupy

    return cupy


def numpy_only_compatibility(numpy_func):
    """
    Apply this decorator to numpy only functions to make them compliant
    with the array-api
    numpy_func: Callable
        signature (*args, **kwargs) -> array like object
    """

    @functools.wraps(numpy_func)
    def func(*args, **kwargs):
        array_args = list(filter(is_array_api_obj, args))
        array_kwargs = list(filter(is_array_api_obj, kwargs.values()))
        xp = array_namespace(*array_args, *array_kwargs)
        numpy_args = []
        for arg in args:
            arg_ = arg
            if is_array_api_obj(arg):
                arg_ = _to_numpy(arg)
            numpy_args.append(arg_)
        numpy_kwargs = {}
        for k, arg in kwargs.items():
            arg_ = arg
            if is_array_api_obj(arg):
                arg_ = _to_numpy(arg)
            numpy_kwargs[k] = arg_

        try:
            devices = set(map(get_device, array_args + array_kwargs))
            if len(devices) != 1:
                raise TypeError(f"Multiple devices found in args: {devices}")
            (device,) = devices
        except TypeError:
            device = get_device(array_args[0])

        return xp.asarray(numpy_func(*numpy_args, **numpy_kwargs), device=device)

    return func


@overload
def to_numpy(x: Any) -> numpy.ndarray: ...
@overload
def to_numpy(*xs: Any) -> tuple[numpy.ndarray, ...]: ...
def to_numpy(*xs):
    ret = tuple([_to_numpy(x) for x in xs])
    if len(xs) == 1:
        return ret[0]
    else:
        return ret


def _to_numpy(x) -> numpy.ndarray:
    if is_numpy_array(x):
        return x
    elif is_cupy_array(x):
        return x.get()
    elif is_torch_array(x):
        return x.cpu().numpy()
    else:
        return numpy.asarray(x, copy=True, device="cpu")


def _is_numpy_namespace(xp):
    return xp.__name__ in ("numpy", "array_api_compat.numpy")


def _is_cupy_namespace(xp):
    return xp.__name__ in ("cupy", "array_api_compat.cupy")


def _is_torch_namespace(xp):
    return xp.__name__ in ("torch", "array_api_compat.torch")


def get_prefered_namespace_device(
    xp=None,
    device=None,
    gpu=None,
):
    if xp is not None:
        if device is not None:
            return xp, device
        if _is_torch_namespace(xp):
            xp = get_torch()
            if gpu is None:
                device = "cuda" if spfluo.has_torch_cuda() else "cpu"
            elif gpu:
                device = "cuda"
            else:
                device = "cpu"
        elif _is_cupy_namespace(xp):
            xp = get_cupy()
            if gpu is False:
                raise RuntimeError(f"{xp} cannot create non cuda arrays")
            device = None
        elif _is_numpy_namespace(xp):
            xp = numpy
            if gpu is True:
                raise RuntimeError(f"{xp} cannot create gpu arrays")
            device = "cpu"
        else:
            raise RuntimeError(
                f"{xp} not supported. Must be one of torch, cupy or {numpy}"
            )
    else:
        if device is not None:
            raise RuntimeError("device provided but xp not provided.")
        if gpu is None:
            if spfluo.has_torch_cuda():
                xp = get_torch()
                device = "cuda"
            elif spfluo.has_cupy():
                xp = get_cupy()
                device = None
            elif spfluo.has_torch():
                xp = get_torch()
                device = "cpu"
            else:
                xp = numpy
                device = "cpu"
        elif gpu:
            if spfluo.has_torch_cuda():
                xp = get_torch()
                device = "cuda"
            elif spfluo.has_cupy():
                xp = get_cupy()
                device = None
            else:
                raise RuntimeError("GPU asked but no backend found")
        else:
            if spfluo.has_torch():
                xp = get_torch()
                device = "cpu"
            else:
                xp = numpy
                device = "cpu"
    return xp, device


def median(a, axis=None, xp=None):
    xp = array_namespace(a) if xp is None else xp
    if axis is None:
        a = xp.reshape(a, (-1,))
        axis = 0
    N = a.shape[axis]
    if N % 2 == 1:
        res = xp.take(
            xp.sort(a, axis=axis), xp.asarray((a.shape[axis] - 1) // 2), axis=axis
        )
    else:
        res = (
            xp.take(xp.sort(a, axis=axis), xp.asarray(a.shape[axis] // 2), axis=axis)
            + xp.take(
                xp.sort(a, axis=axis), xp.asarray(a.shape[axis] // 2 - 1), axis=axis
            )
        ) / 2
    res = xp.astype(res, xp.float64)
    return res


__all__ = [array_namespace, is_array_api_obj, to_device, to_numpy, get_device]

if TYPE_CHECKING:
    __all__ = [
        array_namespace,
        is_array_api_obj,
        to_device,
        to_numpy,
        get_device,
    ]



from typing import Tuple

import numpy as np
import torch
import torch.nn.functional as F


def affine_transform_batched_multichannel_pytorch(
    input: torch.Tensor,
    matrix: torch.Tensor,
    offset=0.0,
    output_shape=None,
    output=None,
    order=1,
    mode="zeros",
    cval=0.0,
    prefilter=True,
) -> torch.Tensor:
    """Rotate the volume according to the transform matrix.
    Matches the `scipy.ndimage.affine_transform` function at best with the
    `torch.nn.functional.grid_sample` function

    Args:
        input (torch.Tensor): 3D images of shape (N, C, D, H, W)
        matrix (torch.Tensor)
            transform matrices of shape (N, 3), (N,3,3), (N,4,4) or (N,3,4).
            Should be of floating point dtype
        offset (float or torch.Tensor): offset of the grid.
        output_shape (tuple): shape of the output.
        output: not implemented
        order (int): must be 1. Only linear interpolation is implemented.
        mode (str): Points outside the boundaries of the input are filled
            according to the given mode
            Only ``'constant'``, ``'nearest'``, ``'reflect'`` are implemented.
        cval (float): cannot be different than 0.0
        prefilter (bool): not implemented

    Returns:
        torch.Tensor: Rotated volumes of shape (N, C, D, H, W)
    """
    N, C, D, H, W = input.size()
    dtype = input.dtype
    (device,) = set((input.device, matrix.device))
    tensor_kwargs = dict(device=device, dtype=dtype)

    if isinstance(offset, float):
        tvec = torch.tensor([offset, offset, offset], **tensor_kwargs).expand(N, 3)
    elif offset.shape == (3,):
        tvec = torch.as_tensor(offset, **tensor_kwargs).expand(N, 3)
    elif offset.shape == (N, 3):
        tvec = torch.as_tensor(offset, **tensor_kwargs)
    else:
        raise ValueError(
            "Offset should be a float, a sequence of size 3 or a tensor of size (N,3)."
        )

    if matrix.size() == torch.Size([N, 3, 3]):
        rotMat = matrix
    elif matrix.size() == torch.Size([N, 3]):
        rotMat = torch.stack([torch.diag(matrix[i]) for i in range(N)])
    elif matrix.size() == torch.Size([N, 4, 4]) or matrix.size() == torch.Size(
        [N, 3, 4]
    ):
        rotMat = matrix[:, :3, :3]
        tvec = matrix[:, :3, 3]
    else:
        raise ValueError(
            "Matrix should be a tensor of shape"
            f"{(N,3)}, {(N,3,3)}, {(N,4,4)} or {(N,3,4)}."
            f"Found matrix of shape {matrix.size()}"
        )

    if output_shape is None:
        output_shape = (D, H, W)

    if output is not None:
        raise NotImplementedError()

    if order > 1:
        raise NotImplementedError("order > 1 is not implemented")

    pytorch_modes = {"constant": "zeros", "nearest": "border", "reflect": "reflection"}
    if mode not in pytorch_modes:
        raise NotImplementedError(f"Only {pytorch_modes.keys()} are available")
    pt_mode = pytorch_modes[mode]

    if cval != 0:
        raise NotImplementedError()

    if order > 1 and prefilter:
        raise NotImplementedError()

    return _affine_transform(
        input, rotMat.type(input.dtype), tvec, output_shape, pt_mode, **tensor_kwargs
    )


def _affine_transform(input, rotMat, tvec, output_shape, mode, **tensor_kwargs):
    output_shape = list(output_shape)
    rotated_vol = input.new_empty(list(input.shape[:2]) + output_shape)
    grid = torch.stack(
        torch.meshgrid(
            [torch.linspace(0, d - 1, steps=d, **tensor_kwargs) for d in output_shape],
            indexing="ij",
        ),
        dim=-1,
    )
    c = torch.tensor([0 for d in output_shape], **tensor_kwargs)
    input_shape = torch.as_tensor(input.shape[2:], **tensor_kwargs)
    grid = (
        (rotMat[:, None, None, None] @ ((grid - c)[None, ..., None]))[..., 0]
        + c
        + tvec[:, None, None, None, :]
    )
    grid = -1 + 1 / input_shape + 2 * grid / input_shape
    rotated_vol = F.grid_sample(
        input,
        grid[:, :, :, :, [2, 1, 0]],
        mode="bilinear",
        align_corners=False,
        padding_mode=mode,
    )
    return rotated_vol


def pad_to_size(volume: torch.Tensor, output_size: torch.Size) -> torch.Tensor:
    output_size = torch.as_tensor(output_size)
    pad_size = torch.ceil((output_size - torch.as_tensor(volume.size())) / 2)

    padding = tuple(
        np.asarray(
            [[max(pad_size[i], 0), max(pad_size[i], 0)] for i in range(len(pad_size))],
            dtype=int,
        )
        .flatten()[::-1]
        .tolist()
    )
    output_volume = F.pad(volume, padding)

    shift = (torch.as_tensor(output_volume.size()) - output_size) / 2
    slices = [
        slice(int(np.ceil(shift[i])), -int(np.floor(shift[i])))
        if shift[i] > 0 and np.floor(shift[i]) > 0
        else slice(int(np.ceil(shift[i])), None)
        if shift[i] > 0
        else slice(None, None)
        for i in range(len(shift))
    ]

    return output_volume[tuple(slices)]


def hann_window(shape: Tuple[int], **kwargs) -> torch.Tensor:
    """Computes N dimensional Hann window.

    Args:
        shape: shape of the final window
        kwargs: keyword arguments for torch.hann_window function
    Returns:
         Hann window of the shape asked
    """
    windows = [torch.hann_window(s, **kwargs) for s in shape]
    view = [1] * len(shape)
    hw = torch.ones(shape, device=windows[0].device, dtype=windows[0].dtype)
    for i in range(len(windows)):
        view_ = list(view)
        view_[i] = -1
        hw *= windows[i].view(tuple(view_))
    return hw


def normalize_patches(patches: torch.Tensor) -> torch.Tensor:
    """Normalize N patches by computing min/max for each patch
    Params: patches (torch.Tensor) of shape (N, ...)
    Returns: normalized_patches (torch.Tensor) of shape (N, ...)
    """
    N = patches.size(0)
    patch_shape = patches.shape[1:]
    flatten_patches = patches.view(N, -1)
    min_patch, _ = flatten_patches.min(dim=1)
    max_patch, _ = flatten_patches.max(dim=1)
    min_patch = min_patch.view(tuple([N] + [1] * len(patch_shape)))
    max_patch = max_patch.view(tuple([N] + [1] * len(patch_shape)))
    normalized_patches = (patches - min_patch) / (max_patch - min_patch)

    return normalized_patches



import argparse
import logging

logging.basicConfig()
spfluo_logger = logging.getLogger("spfluo")

base_parser = argparse.ArgumentParser(
    "spfluo cli tool",
    add_help=False,
)
base_parser.add_argument("--debug", action="store_true")


def set_logging_level(args: argparse.Namespace):
    debug_arg, _ = base_parser.parse_known_args(namespace=args)
    if debug_arg.debug:
        spfluo_logger.setLevel(logging.DEBUG)



import argparse

from ..viewers import show_points


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("file", type=str, help="path to the image to visualize")
    parser.add_argument("--coords", type=str, required=True)
    parser.add_argument(
        "--scale", type=float, default=None, nargs="+", help="Voxel size (ZYX)"
    )
    args = parser.parse_args()

    show_points(args.file, args.coords, args.scale)


if __name__ == "__main__":
    main()
